from __future__ import annotations
from dataclasses import dataclass
from typing import cast, Optional, TypeVar
from typing import Tuple

from relationalai.early_access.metamodel import builtins, ir, factory as f, helpers, types, visitor
from relationalai.early_access.metamodel.compiler import Pass, group_tasks
from relationalai.early_access.metamodel.util import OrderedSet, ordered_set, NameCache
from relationalai.early_access.metamodel import dependency
from relationalai.early_access.metamodel.util import FrozenOrderedSet, filter_by_type
from relationalai.early_access.metamodel.typer.typer import to_type

class Flatten(Pass):
    """
    Traverses the model's root to flatten it as much as possible. The result of this pass is
    a Logical root where all nested tasks that represent a rule in Rel are extraced to the
    top level.

    - nested logical with updates becomes a top-level logical (a rule)

    From:
        Logical
            Logical
                lookup1   <- scope is spread
                Logical
                    lookup2
                    derive foo
                Logical
                    lookup3
                    derive bar
    To:
        Logical
            Logical
                lookup1
                lookup2
                derive foo
            Logical
                lookup1
                lookup3
                derive bar

    - nested logical with aggregates becomes a top-level logical (a rule representing an aggregation)

    From:
        Logical
            Logical
                lookup1
                Logical
                    lookup2
                    aggregate1
                Logical
                    lookup3
                    aggregate2
                output
    To:
        Logical
            Logical
                lookup1
                lookup2
                aggregate1
                derive tmp1
            Logical
                lookup1
                lookup3
                aggregate2
                derive tmp2
            Logical
                lookup1
                lookup tmp1
                lookup tmp2
                output

    - a union becomes a top-level logical for each branch, writing into a temporary relation,
    and a lookup from that relation.

    From:
        Logical
            Logical
                Union
                    Logical
                        lookup1
                    Logical
                        lookup2
                output
    To:
        Logical
            Logical
                lookup1
                derive tmp1
            Logical
                lookup2
                derive tmp1
            Logical
                lookup tmp1
                output

    - a match becomes a top-level logical for each branch, each writing into its own temporary
    relation and a lookup from the last relation. The top-level logical for a branch derives
    into the temporary relation negating the previous branch:

    From:
        Logical
            Logical
                Match
                    Logical
                        lookup1
                    Logical
                        lookup2
                output
    To:
        Logical
            Logical
                lookup1
                derive tmp1
            Logical
                Union            <- tmp1() or (not temp1() and lookup2())
                    lookup tmp1
                    Logical
                        Not
                            lookup tmp1
                        lookup2
                        derive tmp2
            Logical
                lookup tmp2
                output
    """

    def __init__(self, handle_outputs: bool=True, handle_union: bool=True):
        super().__init__()
        self.name_cache = NameCache()
        self._handle_union = handle_union
        self._handle_outputs = handle_outputs

    def create_task_name(self, task: ir.Task, prefix: Optional[str]=None) -> str:
        prefix = prefix if prefix else f"_{task.kind}"
        return self.name_cache.get_name(task.id, prefix, start_from_one=True)

    #--------------------------------------------------
    # Public API
    #--------------------------------------------------
    def rewrite(self, model: ir.Model, options:dict={}) -> ir.Model:
        # create the context
        ctx = Flatten.Context(model, options)

        # rewrite the root
        result = self.handle(model.root, ctx)

        # the new body contains the extracted top level logicals and maybe the rewritten root
        body = ctx.rewrite_ctx.top_level if result.replacement is None else ctx.rewrite_ctx.top_level + [result.replacement]

        # create the new model, updating relations and root
        return ir.Model(
            model.engines,
            OrderedSet.from_iterable(model.relations).update(ctx.rewrite_ctx.relations).frozen(),
            model.types,
            ir.Logical(model.root.engine, tuple(), tuple(body))
        )

    #--------------------------------------------------
    # IR handlers
    #--------------------------------------------------

    class Context():
        def __init__(self, model: ir.Model, options: dict):
            self.rewrite_ctx = helpers.RewriteContext()
            self.info = dependency.analyze(model.root)
            self.options = options
            # when we extract a nested logical (due to agg, match, union, etc), we also
            # extract together its dependencies, and we leave behind a different nested
            # logical as a reference, with a lookup to the extracted relation. This dict
            # keeps a map from the reference logical id to the set of tasks that where
            # extracted for it. This is used to optimize the resulting code, since we can
            # assume that the reference logical already includes those tasks.
            self.included: dict[int, OrderedSet[ir.Task]] = dict()
            # These are extra tasks to be added to the body of a logical when it is
            # extracted as a top-level logical. For example, when handling a match, we add
            # a couple of tasks here in case the branches have updates or aggregations that
            # cause the branch to be extracted, so that they include these tasks.
            self.extra_tasks = []


    @dataclass
    class HandleResult():
        replacement: Optional[ir.Task]

    def handle(self, task: ir.Task, ctx: Context) -> Flatten.HandleResult:
        if isinstance(task, ir.Logical):
            return self.handle_logical(task, ctx)
        elif isinstance(task, ir.Union) and self._handle_union:
            return self.handle_union(task, ctx)
        elif isinstance(task, ir.Match):
            return self.handle_match(task, ctx)
        elif isinstance(task, ir.Require):
            return self.handle_require(task, ctx)
        else:
            return Flatten.HandleResult(task)

    def handle_logical(self, task: ir.Logical, ctx: Context):
        # keep track of what's the result of handling nested composites
        composites = group_tasks(task.body, {
            "composites": helpers.COMPOSITES
        })["composites"]
        all_composites_removed = len(composites) > 0

        # recursively handle children, collecting the replacements in the body
        body:OrderedSet[ir.Task] = ordered_set()
        for child in task.body:
            result = self.handle(child, ctx)
            if result.replacement is not None:
                ctx.info.replaced(child, result.replacement)
                extend_body(body, result.replacement)
                # nested composite was not completely removed
                if child in composites:
                    all_composites_removed = False

        # all children were extracted or all composites were removed without any effects
        # left and no outputs (so no way for outer dependencies), drop this logical
        if not body or (all_composites_removed and not any([isinstance(t, helpers.EFFECTS) for t in body]) and not ctx.info.task_outputs(task)):
            return Flatten.HandleResult(None)

        # now process the rewritten body
        groups = group_tasks(body.get_list(), {
            "outputs": ir.Output,
            "updates": ir.Update,
            "aggregates": ir.Aggregate,
            "ranks": ir.Rank,
        })

        # if there are outputs, adjust them (depending on the config for wide vs gnf)
        if groups["outputs"] and self._handle_outputs:
            return self.adjust_outputs(task, body, groups, ctx)

        # if there are updates, extract as a new top level rule
        if groups["updates"]:
            # add task dependencies to the body
            body.prefix(ctx.info.task_dependencies(task))
            # potentially add context extra tasks
            body.update(ctx.extra_tasks)
            ctx.rewrite_ctx.top_level.append(ir.Logical(task.engine, task.hoisted, tuple(body), task.annotations))
            return Flatten.HandleResult(None)

        if groups["aggregates"]:
            if len(groups["aggregates"]) > 1:
                # stop rewritting as we don't know how to handle this yet
                return Flatten.HandleResult(task)

            # there must be only one
            agg = cast(ir.Aggregate, groups["aggregates"].some())

            # add agg dependencies to the body
            body.prefix(ctx.info.task_dependencies(agg))

            # extract a new logical for the aggregate, exposing aggregate group-by and results
            exposed_vars = OrderedSet.from_iterable(list(agg.group) + helpers.aggregate_outputs(agg))
            name = self.create_task_name(agg)
            connection = helpers.extract(agg, body, exposed_vars.get_list(), ctx.rewrite_ctx, name)

            # return a reference to the connection relation
            reference = f.logical([f.lookup(connection, exposed_vars.get_list())], task.hoisted)
            return Flatten.HandleResult(reference)

        if groups["ranks"]:
            if len(groups["ranks"]) > 1:
                # stop rewritting as we don't know how to handle this yet
                return Flatten.HandleResult(task)

            # there must be only one
            rank = cast(ir.Rank, groups["ranks"].some())

            # add rank dependencies to the body
            body.prefix(ctx.info.task_dependencies(rank))
            # for rank, we sort by the args, but the result includes the keys to preserve bag semantics.
            exposed_vars_raw = list(rank.projection) + list(rank.group) + list(rank.args) +[rank.result]
            # deduplicate vars
            exposed_vars = OrderedSet.from_iterable(exposed_vars_raw)
            name = self.create_task_name(rank)
            connection = helpers.extract(rank, body, exposed_vars.get_list(), ctx.rewrite_ctx, name)

            # return a reference to the connection relation
            reference = f.logical([f.lookup(connection, exposed_vars.get_list())], task.hoisted)
            ctx.included[reference.id] = body
            return Flatten.HandleResult(reference)

        return Flatten.HandleResult(ir.Logical(task.engine, task.hoisted, tuple(body)))

    def handle_match(self, match: ir.Match, ctx: Context):
        # TODO: how to deal with malformed input like this?
        if not match.tasks:
            return Flatten.HandleResult(match)

        body = ctx.info.task_dependencies(match)
        exposed_vars = self.compute_exposed_vars(match, match.tasks, ctx)

        # negation len is the number of wildcards when negating a reference
        outputs = ctx.info.task_outputs(match)
        negation_len = len(outputs) if outputs else 0

        # collect the references to branch rules so that we union in the end
        references = []

        # a negated reference to the previous branch (e.g. "not branch1(...)")
        negated_reference = None

        # number of negated references currently in context (e.g. in the 3rd branch there
        # will be 2 negated references, "not branch1(...) and not branch2(...)")
        negated_references = 0
        for branch in match.tasks:
            # if there's a previous connection, add the negation of it to extra tasks
            if negated_reference:
                ctx.extra_tasks.append(negated_reference)
                negated_references += 1

            # the connection relation to be used for this branch
            name = self.create_task_name(branch, "_match")
            relation = helpers.create_connection_relation(branch, exposed_vars, ctx.rewrite_ctx, name)
            branch_update = f.derive(relation, exposed_vars)
            ctx.extra_tasks.append(branch_update)

            # handle the branch
            result = self.handle(branch, ctx)

            # if there's no replacement, the branch was already extracted (e.g. by an update),
            # otherwise, we need to extract the branch as a rule writing into the connection
            if result.replacement:
                ctx.info.replaced(branch, result.replacement)

                branch_body: OrderedSet[ir.Task] = OrderedSet.from_iterable(body)
                extend_body(branch_body, result.replacement)
                # potentially add context extra tasks (negations of previous branches + derive into this branch)
                branch_body.update(ctx.extra_tasks)
                # extract the body
                ctx.rewrite_ctx.top_level.append(ir.Logical(match.engine, tuple(), tuple(branch_body)))

            # cleanup the branch update from context (but keep the negations)
            ctx.extra_tasks.pop()

            # record a reference to the branch connection and adjust the negated_reference
            reference = f.lookup(relation, exposed_vars)
            negated_reference = negate(reference, negation_len)
            references.append(reference)

        # we accumulate negations, in the end we remove them all
        for x in range(negated_references):
            ctx.extra_tasks.pop()

        # return a union of the references to the branches
        return Flatten.HandleResult(f.union(references, match.hoisted))


    def handle_union(self, union: ir.Union, ctx: Context):
        # TODO: how to deal with malformed input like this?
        if not union.tasks:
            return Flatten.HandleResult(union)

        body = ctx.info.task_dependencies(union)
        exposed_vars = self.compute_exposed_vars(union, union.tasks, ctx)

        name = self.create_task_name(union)
        relation = helpers.create_connection_relation(union, exposed_vars, ctx.rewrite_ctx, name)
        for branch in union.tasks:

            # add an update to the context in case the branch is extracted
            branch_update = f.derive(relation, exposed_vars)
            ctx.extra_tasks.append(branch_update)

            # handle the branch
            result = self.handle(branch, ctx)

            # cleanup the branch update from context
            ctx.extra_tasks.pop()

            # if there's no replacement, the branch was already extracted (e.g. by an update),
            # otherwise, we need to extract the branch as a rule writing into the connection
            if result.replacement:
                # the branch has some replacement, we have to extract it as a top-level rule
                ctx.info.replaced(branch, result.replacement)
                branch_body: OrderedSet[ir.Task] = OrderedSet.from_iterable(body)
                extend_body(branch_body, result.replacement)
                branch_body.add(branch_update)
                # extract the body
                ctx.rewrite_ctx.top_level.append(ir.Logical(union.engine, tuple(), tuple(branch_body)))

        # return a reference to the connection relation
        reference = f.logical([f.lookup(relation, exposed_vars)], exposed_vars)
        return Flatten.HandleResult(reference)


    def compute_exposed_vars(self, task: ir.Task, branches: Tuple[ir.Task, ...], ctx: Context) -> list:
        """ Helper to compute the set of exposed vars for handle_match and handle_union. """
        exposed_vars = set_union(ctx.info.task_inputs(task), ctx.info.task_outputs(task))
        if exposed_vars:
            return exposed_vars
        # no exposed vars, so we try to use the common vars across all branches as context
        common_vars = helpers.collect_vars(branches[0])
        for branch in branches:
            branch_vars = helpers.collect_vars(branch)
            for v in common_vars:
                if v not in branch_vars:
                    common_vars.remove(v)
        # TODO - we should raise an error here if the set is empty
        return common_vars.get_list()


    def handle_require(self, req: ir.Require, ctx: Context):
        # only extract the domain if it is a somewhat complex Logical and there's more than
        # one check, otherwise insert it straight into all checks
        domain = req.domain
        if len(req.checks) > 1 and isinstance(domain, ir.Logical) and len(domain.body) > 1:
            body = OrderedSet.from_iterable(domain.body)
            vars = helpers.hoisted_vars(domain.hoisted)
            name = self.create_task_name(req)
            connection = helpers.extract(req, body, vars, ctx.rewrite_ctx, name)
            domain = f.logical([f.lookup(connection, vars)], vars)

        for check in req.checks:
            # only generate logic for checks that have errors
            if check.error:
                handled_check_result = self.handle(check.check, ctx)
                if handled_check_result.replacement:
                    body = ordered_set()
                    body.add(domain)
                    body.add(ir.Not(req.engine, handled_check_result.replacement))
                    if (isinstance(check.error, ir.Logical)):
                        body.update(check.error.body)
                    else:
                        # this is more general but may trip the current splinter
                        body.add(check.error)
                    ctx.rewrite_ctx.top_level.append(ir.Logical(req.engine, tuple(), tuple(body)))

        # currently we just drop the Require, but we should keep it here and link the
        # extracted logicals to it
        return Flatten.HandleResult(None)


    #--------------------------------------------------
    # GNF vs wide output support
    #--------------------------------------------------
    def adjust_outputs(self, task: ir.Logical, body: OrderedSet[ir.Task], groups: dict[str, OrderedSet[ir.Task]], ctx: Context):

        # for wide outputs, only adjust the output task to include the keys.
        if ctx.options.get("wide_outputs", False):
            for output in groups["outputs"]:
                assert(isinstance(output, ir.Output))
                if output.keys:
                    body.remove(output)
                    body.add(self.rewrite_wide_output(output))
            # self.remove_subsumptions(body, ctx)
            return Flatten.HandleResult(ir.Logical(task.engine, task.hoisted, tuple(body), task.annotations))

        # for GNF outputs we need to generate a rule for each "column" in the output
        else:
            # first split outputs in potentially multiple outputs, one for each "column"
            for output in groups["outputs"]:
                assert(isinstance(output, ir.Output))
                if output.keys:
                    # we will replace the output bellow,
                    body.remove(output)

                    is_export = builtins.export_annotation in output.annotations
                    # generate an output for each "column"
                    for idx, alias in enumerate(output.aliases):
                        # output looks like def output(:cols, :col000, key0, key1, value):
                        aliases = []
                        if not is_export:
                            aliases.append(("cols", f.literal("cols", types.Symbol)))
                        aliases.append(("col", f.literal(f"col{idx:03}", types.Symbol)))
                        for key in output.keys:
                            aliases.append((f"key_{key.name}_{idx}", key))
                        aliases.append(alias)
                        body.add(ir.Output(
                            output.engine,
                            FrozenOrderedSet.from_iterable(aliases),
                            output.keys,
                            output.annotations
                        ))

            # analyse the resulting logical to be able to pull dependencies
            logical = ir.Logical(task.engine, task.hoisted, tuple(body), task.annotations)
            info = dependency.analyze(logical)

            # now extract a logical for each output, bringing together its dependencies
            for output in filter_by_type(body, ir.Output):
                deps = info.task_dependencies(output)
                # TODO: verify safety of doing this
                # self.remove_subsumptions(deps, ctx)
                deps.add(output)
                ctx.rewrite_ctx.top_level.append(ir.Logical(task.engine, tuple(), tuple(deps)))

            return Flatten.HandleResult(None)


    def remove_subsumptions(self, body:OrderedSet[ir.Task], ctx: Context):
        # remove from the body all the tasks that are subsumed by some other task in the set;
        # this can be done because some tasks are references to extracted nested logical that
        # contain filters they dependend on, so we don't need those filters here if the
        # reference is present.
        for logical in filter_by_type(body, ir.Logical):
            if logical.id in ctx.included:
                # if the logical id is included, it means it's a reference to an extracted
                # rule, so remove all other items in the body that are already included in
                # the body referenced by it
                for item in body:
                    if item in ctx.included[logical.id]:
                        body.remove(item)


    def rewrite_wide_output(self, output: ir.Output):
        assert(output.keys)

        # only prefix keys that are not already in the output
        prefix_keys = []
        for key in output.keys:
            if all([val is not key for _, val in output.aliases]):
                prefix_keys.append(key)

        aliases: OrderedSet[Tuple[str, ir.Value]] = ordered_set()
        # add the keys to the output
        for key in prefix_keys:
            aliases.add((key.name, key))

        # add the remaining args, unless it is already a key
        for name, val in output.aliases:
            if not isinstance(val, ir.Var) or val not in prefix_keys:
                aliases.add((name, val))

        # TODO - we are assuming that the Rel compiler will translate nullable lookups
        # properly, returning a `Missing` if necessary, like this:
        # (nested_192(_adult, _adult_name) or (not nested_192(_adult, _) and _adult_name = Missing)) and
        return ir.Output(
            output.engine,
            aliases.frozen(),
            output.keys,
            output.annotations
        )

        # TODO: in the rel compiler, see if we can do this outer join
        # 1. number of keys
        # 2. each relation
        # 3. each variable, starting with the keys
        # 4. tag output with @arrow

        # @arrow def output(_book, _book_title, _author_name):
        #   rel_primitive_outer_join(#1, book_title, author_name, _book, _book_title, _author_name)
        # def output(p, n, c):
        #     rel_primitive_outer_join(#1, name, coolness, p, n, c)

#--------------------------------------------------
# Helpers
#--------------------------------------------------

T = TypeVar('T')
def set_union(s1: Optional[OrderedSet[T]], s2: Optional[OrderedSet[T]]) -> list:
    """ Return the union of s1 | s2, accounting for when any is None. """
    if s1 and s2:
        return (s1 | s2).get_list()
    if s1:
        return s1.get_list()
    if s2:
        return s2.get_list()
    return []

def extractable(t: ir.Task):
    """
    Whether this task is a Logical that will be extracted as a top level by this
    pass, because it has an aggregation, effects, match, union, etc.
    """
    extractable_types = (ir.Update, ir.Aggregate, ir.Match, ir.Union, ir.Rank)
    return isinstance(t, ir.Logical) and len(visitor.collect_by_type(extractable_types, t)) > 0

def extractables(composites: OrderedSet[ir.Task]):
    """ Filter the set of composites, keeping only the extractable ones. """
    return list(filter(extractable, composites))

def negate(lookup: ir.Lookup, values: int):
    """
    Return a negation of this reference, where the last `values` arguments are to
    be replaced by wildcards (i.e. len(reference.args) - values are keys so they need
    to be bound in the Not.)
    """
    args = []
    i = 0
    last = len(lookup.args) - values
    for arg in lookup.args:
        args.append(f.wild(to_type(arg))) if i >= last else args.append(arg)
        i += 1

    return ir.Not(lookup.engine, f.lookup(lookup.relation, args))

def merge_var_list(vars: list[ir.Var], hoisted: tuple[ir.VarOrDefault, ...]) -> list[ir.VarOrDefault]:
    """ Merge vars and hoisted, making sure that hoisted vars have precedence since they may have defaults. """
    r = []
    hoisted_vars = helpers.hoisted_vars(hoisted)
    for v in vars:
        if v not in hoisted_vars:
            r.append(v)
    r.extend(hoisted)
    return r

def extend_body(body: OrderedSet[ir.Task], extra: ir.Task):
    """ Add the extra task to the body, but if the extra is a simple logical, just
    inline its subtasks. """
    if isinstance(extra, ir.Logical):
        if extra.hoisted:
            # hoists, remove things that are already in the body to avoid duplicates
            logical_body = []
            for t in extra.body:
                if t not in body:
                    logical_body.append(t)
            if len(logical_body) == len(extra.body):
                # no duplicates
                body.add(extra)
            else:
                # some duplicate, remove them
                body.add(ir.Logical(
                    extra.engine,
                    extra.hoisted,
                    tuple(logical_body)
                ))
        else:
            # no hoists, just inline
            body.update(extra.body)
    else:
        body.add(extra)



from relationalai.early_access.dsl.bindings.common import BindableColumn, AbstractBindableTable
from relationalai.early_access.dsl.snow.common import ColumnMetadata, _map_rai_type


#=
# Bindable classes and interfaces.
#=

class BindableSnowflakeColumn(BindableColumn):
    _metadata: ColumnMetadata

    def __init__(self, metadata: ColumnMetadata, table: 'SnowflakeTable', model):
        col_name = metadata.name
        col_type = _map_rai_type(metadata)
        super().__init__(col_name, col_type, table, model)
        self._metadata = metadata

    @property
    def metadata(self):
        return self._metadata

    def decimal_scale(self):
        return self._metadata.numeric_scale

    def decimal_size(self):
        precision = self._metadata.numeric_precision
        if precision is not None:
            if 1 <= precision <= 18:
                return 64
            elif 18 < precision <= 38:
                return 128
            raise ValueError(f'Precision {precision} is not supported (max: 38)')
        return precision

    def __repr__(self):
        return f"Snowflake:{super().__repr__()}"


class SnowflakeTable(AbstractBindableTable[BindableSnowflakeColumn]):

    def __init__(self, fqn: str, model):
        self._metadata = model.api().table_metadata(fqn)
        super().__init__(fqn, model, self._metadata.foreign_keys)
        self._initialize(model)

    def _initialize(self, model):
        self._model = model
        self._cols = {col.name: BindableSnowflakeColumn(col, self, model) for col in self._metadata.columns}
        self._process_foreign_keys()
        # initialize the table so that the graph index can be updated
        self._initialize_qb_table()

    def _initialize_qb_table(self):
        from relationalai.early_access.builder.snowflake import Table as QBTable
        self._qb_table = QBTable(self._table)
        QBTable._used_sources.add(self._qb_table)

    def __str__(self):
        # returns the name of the table, as well as the columns and their types
        return self.physical_name() + ':\n' + '\n'.join(
            [f' {col.metadata.name} {col.metadata.datatype}' for _, col in self._cols.items()]
        ) + '\n' + '\n'.join(
            [f' {fk.source_columns} -> {fk.target_columns}' for fk in self._foreign_keys]
        )

    def physical_name(self):
        # physical relation name is always in the form of `{database}_{schema}_{table}
        return f"{self._table.lower()}".replace('.', '_')



from __future__ import annotations

__all__ = [
    "SourceInfo",
    "LqpNode",
    "Declaration",
    "Def",
    "Loop",
    "Abstraction",
    "Formula",
    "Exists",
    "Reduce",
    "Conjunction",
    "Disjunction",
    "Not",
    "FFI",
    "Atom",
    "Pragma",
    "Primitive",
    "RelAtom",
    "Cast",
    "Var",
    "UInt128",
    "PrimitiveValue",
    "Constant",
    "Term",
    "Specialized",
    "RelTerm",
    "Attribute",
    "RelationId",
    "RelValueType",
    "PrimitiveType",
    "RelType",
    "FragmentId",
    "Fragment",
    "Define",
    "Write",
    "Output",
    "Read",
    "Epoch",
    "Transaction",
    "DebugInfo",
    "ir_to_proto",
    "convert_transaction",
    "validate_lqp"
]

from lqp.ir import (
    SourceInfo,
    LqpNode,
    Declaration,
    Def,
    Loop,
    Abstraction,
    Formula,
    Exists,
    Reduce,
    Conjunction,
    Disjunction,
    Not,
    FFI,
    Atom,
    Pragma,
    Primitive,
    RelAtom,
    Cast,
    Var,
    UInt128,
    PrimitiveValue,
    Constant,
    Term,
    Specialized,
    RelTerm,
    Attribute,
    RelationId,
    RelValueType,
    PrimitiveType,
    RelType,
    FragmentId,
    Fragment,
    Define,
    Write,
    Output,
    Read,
    Epoch,
    Transaction,
    DebugInfo
)

from lqp.emit import (
    ir_to_proto,
    convert_transaction
)

from lqp.validator import (
    validate_lqp
)



import re
from typing import Optional

from relationalai.early_access.dsl.core.types.standard import standard_value_types
from relationalai.early_access.dsl.core.utils import get_values_from_keys, camel_to_snake
from relationalai.util.graph import topological_sort
from relationalai.early_access.dsl.ontologies.models import Model
from relationalai.early_access.dsl.ontologies.relationships import Relationship
from relationalai.early_access.dsl.serialize.model import ModelObject


def load_models(json_data) -> list[Model]:
    model_objects = ModelObject.schema().load(json_data, many=True) # type: ignore
    return convert_to_model(model_objects)


def convert_to_model(model_objects: list[ModelObject]) -> list[Model]:
    result = []
    for obj in model_objects:
        m = obj.model
        model = Model(m.name, m.is_primary)

        # Populate ValueTypes
        value_types_map = {vt.name: vt for vt in m.concepts.value_types}
        for vt in m.concepts.value_types:
            if vt.data_type_name not in standard_value_types:
                raise ValueError(f"Unsupported data type '{vt.data_type_name}' in '{m.name}'")
            # todo [AN]: add support for complex Value types with value constraints
            model.value_type(vt.name, standard_value_types[vt.data_type_name])

        # Populate EntityTypes
        # todo [AN]: add support for external UC as a preferred identifier
        entity_types_map = {et.name: et for et in m.concepts.entity_types}
        # First of all do topological sort to init complex entity types at the end
        ordered_concepts = _sort_dependency_graph(m)
        for c in ordered_concepts:
            if c in value_types_map or c in standard_value_types:
                # skip value types
                continue
            if c not in entity_types_map:
                raise ValueError(f"The concept '{c}' was not declared in '{m.name}' but used as concept domain.")
            et = entity_types_map[c]
            model.entity_type(et.name)

        # Populate subtype arrows
        for a in m.subtype_arrows:
            model.subtype_arrow(_concept_lookup(a.end_name, model), [_concept_lookup(a.start_name, model)])

        # Populate relationships
        roles_map = {}
        for rel in m.relationships:
            relationship: Optional[Relationship]
            if rel.is_subtype:
                # Subtype relationships are created when entity subtypes are created.
                relationship = model.lookup_relationship(rel.name)
                if relationship is None:
                    raise Exception(f"Could not find a relationship for '{rel.name}' in '{m.name}'")
            else:
                # if `rel.name` contains `IsIdentifiedBy` then we shouldn't create a new relationship if already exists
                relationship = model.lookup_relationship(rel.name)
                if "IsIdentifiedBy" in rel.name and relationship is not None:
                    _update_role_map(relationship, roles_map)
                    continue

                role_guid_to_role_name = {r.id: r.name if r.name else camel_to_snake(r.role_player_name)
                                          for r in rel.roles}
                relationship = model.relationship()
                for idx, r in enumerate(rel.roles):
                    relationship.role(_concept_lookup(r.role_player_name, model), r.name)
                    relationship.role_at(idx).verbalization(prefix=r.pre_bound_text, postfix=r.post_bound_text)
                # Add readings.
                for red in rel.readings:
                    parts = _get_parts(red.text)
                    args = []
                    for i, r_id in enumerate(red.roles):
                        args.append(relationship._rolemap.get(role_guid_to_role_name[r_id]))
                        if i != len(rel.roles) - 1:
                            args.append(parts[i])
                    relationship.relation(*args)
            _update_role_map(relationship, roles_map)

        # Populate constrains
        for c in m.constraints.mandatory:
            for r in c.roles:
                model.mandatory(roles_map.get(r))
        for c in m.constraints.uniqueness:
            roles = get_values_from_keys(roles_map, c.roles)
            if c.is_preferred_identifier:
                relations = []
                for role in roles:
                    role_player_name = role.player().display()
                    relationship = role.part_of
                    if relationship is None:
                        raise ValueError(f"Role player {role_player_name} is not part of any relationship")

                    # Find the matching relation where the 2nd role is `role`
                    matching_relation = next(
                        (rel for rel in relationship.relations() if rel.reading().roles[1] == role),
                        None
                    )

                    if not matching_relation:
                        rel_name = relationship._name()
                        raise ValueError(f"No matching relation found for role player '{role_player_name}' "
                                         f"in relationship '{rel_name}'")

                    relations.append(matching_relation)

                model.ref_scheme(*relations)
            else:
                model.unique(*roles)
        result.append(model)
    return result

def _update_role_map(relationship, roles_map):
    roles_map.update({r.guid(): r for r in relationship._rolemap.values()})


def _sort_dependency_graph(m):
    nodes = []
    edges = []
    for et in m.concepts.entity_types:
        nodes.append(et.name)
        for d in et.domain:
            edges.append((d, et.name))
    for a in m.subtype_arrows:
        edges.append((a.start_name, a.end_name))
    return topological_sort(nodes, edges)

def _concept_lookup(name, model):
    concept = model.lookup_concept(name)
    if concept is None:
        # sanity check, should never happen
        raise ValueError(
            f"Trying to refer to the concept '{name}' which was not created in model '{model.name}'.")
    return concept

def _get_parts(text):
    pattern = re.compile(r"\{\d+\}(.*?)(?=\{\d+\}|$)")
    matches = pattern.findall(text)
    return [match.strip() for match in matches]




"""
    Elementary IR types.
"""
from . import ir, util
import sys
from typing import cast, Optional
from functools import reduce
from decimal import Decimal as PyDecimal
from datetime import date, datetime

#
# Basic Types
#
Null = ir.ScalarType("Null", util.frozen())
Any = ir.ScalarType("Any", util.frozen())
Hash = ir.ScalarType("Hash", util.frozen())
String = ir.ScalarType("String", util.frozen())
Number = ir.ScalarType("Number", util.frozen())
Int = ir.ScalarType("Int", util.frozen())
Float = ir.ScalarType("Float", util.frozen())
Decimal64 = ir.ScalarType("Decimal64", util.frozen())
Decimal128 = ir.ScalarType("Decimal128", util.frozen())
Bool = ir.ScalarType("Bool", util.frozen())
Binary = ir.ScalarType("Binary", util.frozen()) # 0 or 1
Date = ir.ScalarType("Date", util.frozen())
DateTime = ir.ScalarType("DateTime", util.frozen())
Symbol = ir.ScalarType("Symbol", util.frozen(String))
Sha1 = ir.ScalarType("SHA1", util.frozen())
RowId = ir.ScalarType("RowId", util.frozen(Int))
Enum = ir.ScalarType("Enum", util.frozen())

# a special type that represent an entity that is of a
# single type in a given expression. This allows us to create
# overloads for e.g. Person < Person, but fails Person < Car
EntityTypeVar = ir.ScalarType("EntityTypeVar", util.frozen())

AnyList = ir.ListType(Any)

def is_builtin(t: ir.Type):
    return t in builtin_types

def _compute_builtin_types() -> list[ir.Type]:
    module = sys.modules[__name__]
    types = []
    for name in dir(module):
        builtin = getattr(module, name)
        if isinstance(builtin, ir.Type):
            types.append(builtin)
    return types

# Types that can appear in literal values in the AST. These must be included in all models.
literal_types = util.ordered_set(*[String, Bool, Int, Float, Decimal128, Date, DateTime]).frozen()
py_literal_types = (str, bool, int, float, date, datetime, PyDecimal)

builtin_types = _compute_builtin_types()
builtin_scalar_types_by_name = dict((t.name, t) for t in cast(list[ir.ScalarType], util.filter_by_type(builtin_types, ir.ScalarType)))

def is_any(t: ir.Type) -> bool:
    return t == Any

def is_value_base_type(t: ir.Type) -> bool:
    return isinstance(t, ir.ScalarType) and t in [Hash, String, Number, Int, Float, Decimal64, Decimal128, Bool, Binary, Symbol, Sha1, Date, DateTime]

def is_value_type(t: ir.Type) -> bool:
    if is_value_base_type(t):
        return True
    if isinstance(t, ir.ScalarType):
        t = cast(ir.ScalarType, t)
        if len(t.super_types) == 1:
            # Not sure why we need this cast, but it does shut up the type checker.
            s = cast(ir.Type, t.super_types[0])
            if is_value_type(s):
                return True
    return False

def is_entity_type(t: ir.Type) -> bool:
    return isinstance(t, ir.ScalarType) and not is_value_type(t)

def is_null(t: ir.Type) -> bool:
    return t == Null

def is_abstract_type(t: ir.Type) -> bool:
    if isinstance(t, ir.ScalarType):
        return t in [Any, Number]
    elif isinstance(t, ir.ListType):
        return is_abstract_type(t.element_type)
    elif isinstance(t, ir.TupleType):
        return any(is_abstract_type(t) for t in t.types)
    elif isinstance(t, ir.UnionType):
        return any(is_abstract_type(t) for t in t.types)
    else:
        return False

def union(*types: ir.Type) -> ir.Type:
    if len(types) == 0:
        return Null
    elif len(types) == 1:
        return types[0]

    if any(isinstance(t, ir.UnionType) for t in types):
        def _union_elements(t: ir.Type) -> util.FrozenOrderedSet[ir.Type]:
            if isinstance(t, ir.UnionType):
                return t.types
            else:
                return util.ordered_set(t).frozen()
        types = tuple([u for t in types for u in _union_elements(t) if not is_null(u)])

    # Remove types subsumed by other types.
    types = tuple([t for t in types if not any(is_proper_subtype(t, s) for s in types)])

    # Check again for the simple cases.
    if len(types) == 0:
        return Null
    elif len(types) == 1:
        return types[0]

    return ir.UnionType(util.ordered_set(*types).frozen())

def intersect(*types: ir.Type) -> ir.Type:
    """
    Intersect a list of types.
    """
    return reduce(intersect_binary, types, Any)

def intersect_binary(t1: ir.Type, t2: ir.Type) -> ir.Type:
    if is_null(t1) or is_null(t2):
        return Null
    elif is_any(t1):
        return t2
    elif is_any(t2):
        return t1
    elif is_subtype(t1, t2):
        return t1
    elif is_subtype(t2, t1):
        return t2
    elif isinstance(t1, ir.UnionType):
        return union(*[intersect_binary(s1, t2) for s1 in t1.types])
    elif isinstance(t2, ir.UnionType):
        return union(*[intersect_binary(t1, s2) for s2 in t2.types])
    elif isinstance(t1, ir.ListType) and isinstance(t2, ir.ListType):
        return ir.ListType(intersect_binary(t1.element_type, t2.element_type))
    elif isinstance(t1, ir.TupleType) and isinstance(t2, ir.TupleType):
        if len(t1.types) != len(t2.types):
            return Null
        else:
            return ir.TupleType(tuple([intersect(s1, s2) for s1, s2 in zip(t1.types, t2.types)]))
    else:
        return Null

def matches(t1: ir.Type, t2: ir.Type) -> bool:
    return is_subtype(t1, t2) or is_subtype(t2, t1)

def is_proper_subtype(t1: ir.Type, t2: ir.Type) -> bool:
    return t1 != t2 and is_subtype(t1, t2)

def is_subtype(t1: ir.Type, t2: ir.Type) -> bool:
    """
    Check if t1 is a subtype of t2.
    """
    if isinstance(t1, ir.ScalarType) and t2 == Any:
        # Any is the universal supertype of all scalar types.
        return True
    elif t1 == Null and isinstance(t2, ir.ScalarType):
        # Null is the universal subtype of all scalar types.
        return True
    elif isinstance(t1, ir.ScalarType) and isinstance(t2, ir.ScalarType):
        return t1 == t2 or any(is_subtype(s1, t2) for s1 in t1.super_types)
    elif isinstance(t1, ir.ListType) and isinstance(t2, ir.ListType):
        # Lists are covariant.
        return is_subtype(t1.element_type, t2.element_type)
    elif isinstance(t1, ir.TupleType) and isinstance(t2, ir.TupleType):
        # Tuples are covariant.
        return all(is_subtype(a1, a2) for a1, a2 in zip(t1.types, t2.types))
    elif isinstance(t1, ir.ScalarType) and isinstance(t2, ir.ListType):
        return is_subtype(t1, t2.element_type)
    elif isinstance(t1, ir.TupleType) and isinstance(t2, ir.ListType):
        # list[t] is a supertype of tuple types (t, ...).
        return all(is_subtype(a1, t2.element_type) for a1 in t1.types)
    elif isinstance(t1, ir.UnionType):
        # (t1 | ... | tn) <: t if ti <: t for all i.
        return all(is_subtype(a1, t2) for a1 in t1.types)
    elif isinstance(t2, ir.UnionType):
        # t <: (t1 | ... | tn) if t <: ti for some i.
        return any(is_subtype(t1, a2) for a2 in t2.types)
    else:
        return False

def compute_highest_bound(bound: util.OrderedSet[ir.Type]) -> Optional[ir.Type]:
    """
    Compute the highest bound from a set of types.
    The highest bound is the union of all types that have no proper supertype in the bound.
    """
    if len(bound) == 1:
        return bound[0]
    if len(bound) == 0:
        return None

    # Include only the types that have no proper supertype in the bound.
    ts = []
    for t in bound:
        if not any(is_proper_subtype(t, t2) for t2 in bound):
            ts.append(t)

    if len(ts) == 0:
        return None
    elif len(ts) == 1:
        return ts[0]
    else:
        return ir.UnionType(util.ordered_set(*ts).frozen())

def compute_lowest_bound(bound: util.OrderedSet[ir.Type]) -> Optional[ir.Type]:
    """
    Compute the lowest bound from a set of types.
    The lowest bound is the union of all types that have no proper subtype in the bound.
    """
    if len(bound) == 1:
        return bound[0]
    if len(bound) == 0:
        return None

    # Include only the types that have no proper subtype in the bound.
    ts = []
    for t in bound:
        if not any(is_proper_subtype(t2, t) for t2 in bound):
            ts.append(t)

    if len(ts) == 0:
        return None
    elif len(ts) == 1:
        return ts[0]
    else:
        return ir.UnionType(util.ordered_set(*ts).frozen())



from __future__ import annotations
from typing import Any, Union
from datetime import date, datetime

from .. import builder as b

_Date = Union[b.Producer, date]
_DateTime = Union[b.Producer, datetime]
# TODO support DateTime below as well, but this needs e.g. Rel `datetime_year`

def _make_expr(op: str, *args: Any) -> b.Expression:
    return b.Expression(b.Relationship.builtins[op], *args)

def year(date: _Date) -> b.Expression:
    return _make_expr("date_year", date, b.Integer.ref("res"))

def month(date: _Date) -> b.Expression:
    return _make_expr("date_month", date, b.Integer.ref("res"))

def day(date: _Date) -> b.Expression:
    return _make_expr("date_day", date, b.Integer.ref("res"))

def date_to_datetime(date: _Date, hour: int = 0, minute: int = 0, second: int = 0, millisecond: int = 0, tz: str = "UTC") -> b.Expression:
    _year = year(date)
    _month = month(date)
    _day = day(date)
    return _make_expr("construct_datetime_ms_tz", _year, _month, _day, hour, minute, second, millisecond, tz, b.DateTime.ref("res"))


import json
import re

import xmltodict
from collections import defaultdict
from relationalai.early_access.dsl.adapters.orm.model import ORMEntityType, ORMValueType, ORMRole, ORMSubtypeFact, \
    ORMUniquenessConstraint, ORMExclusionConstraint, ORMMandatoryConstraint, ORMReadingOrderRole, ORMReadingOrder, \
    ExclusiveInclusiveSubtypeFact, SubtypeFact, ExclusiveSubtypeFact, ORMInclusionConstraint, RoleConstraint, \
    ORMRoleValueConstraint, ORMRingConstraint, ORMRingType
from relationalai.util.graph import topological_sort


class ORMParser:

    def __init__(self, orm_file_path):
        with open(orm_file_path) as orm_file:
            data = xmltodict.parse(orm_file.read())

        self._ontology = json.loads(json.dumps(data))

        self._model_name = self._parse_model_name()
        self._entity_types = self._parse_entity_types()
        self._value_types = self._parse_value_types()
        self._object_types = {**self._value_types, **self._entity_types}
        self._roles = self._parse_roles()
        self._fact_type_to_roles = self._parse_fact_type_to_roles()
        self._role_value_constraints = self._parse_role_value_constraints()
        self._internal_uniqueness_constraints, self._external_uniqueness_constraints = self._parse_uniqueness_constraints()
        (self._unique_roles, self._fact_type_to_internal_ucs, self._fact_type_to_complex_ucs,
         self._identifier_fact_type_to_entity_type) = self._process_internal_uniqueness_constraints()
        self._exclusion_constraints = self._parse_exclusion_constraints()
        self._inclusion_constraints = self._parse_inclusion_constraints()
        self._role_constraints = self._parse_role_constraints()
        self._mandatory_constraints = self._parse_mandatory_constraints()
        self._ring_constraints = self._parse_ring_constraints()
        self._subtype_facts = self._parse_subtype_facts()
        self._sorted_subtype_facts = self._sort_subtype_facts()
        self._fact_type_readings = self._parse_fact_types_reading_orders()
        self._mandatory_roles = self._parse_mandatory_roles()

    def model_name(self):
        return self._model_name

    def entity_types(self) -> dict[str, ORMEntityType]:
        return self._entity_types

    def object_types(self):
        return self._object_types

    def value_types(self):
        return self._value_types

    def roles(self):
        return self._roles

    def role_value_constraints(self):
        return self._role_value_constraints

    def subtype_facts(self):
        return self._subtype_facts

    def external_uniqueness_constraints(self):
        return self._external_uniqueness_constraints

    def inclusion_constraints(self):
        return self._inclusion_constraints

    def exclusion_constraints(self):
        return self._exclusion_constraints

    def role_constraints(self):
        return self._role_constraints

    def ring_constraints(self):
        return self._ring_constraints

    def unique_roles(self):
        return self._unique_roles

    def mandatory_roles(self):
        return self._mandatory_roles

    def fact_type_readings(self):
        return self._fact_type_readings

    def fact_type_to_internal_ucs(self):
        return self._fact_type_to_internal_ucs

    def fact_type_to_complex_ucs(self):
        return self._fact_type_to_complex_ucs

    def fact_type_to_roles(self):
        return self._fact_type_to_roles

    def identifier_fact_type_to_entity_type(self):
        return self._identifier_fact_type_to_entity_type

    def sorted_subtype_facts(self):
        return self._sorted_subtype_facts

    def _parse_fact_type_to_roles(self):
        fact_type_data = defaultdict(list)
        for role in self._roles.values():
            relationship_name = role.relationship_name
            fact_type_data[relationship_name].append(role)
        return fact_type_data

    def _parse_model_name(self):
        model_name = self._get_nested(self._ontology, "ormRoot:ORM2", "orm:ORMModel", "@Name")
        return model_name if model_name else "ORMModel"

    def _parse_entity_types(self):
        entity_types = {}
        orm_entity_types = self._get_nested(self._ontology, "ormRoot:ORM2", "orm:ORMModel",
                                            "orm:Objects", "orm:EntityType")
        if orm_entity_types:
            for et in orm_entity_types:
                id = et["@id"]
                name = et['@Name']
                ref_mode = et.get("@_ReferenceMode") or None
                entity_types[id] = ORMEntityType(id, name, ref_mode)
        return entity_types

    def _parse_value_types(self):
        value_types = {}
        data_types = self._parse_data_types()
        orm_value_types = self._get_nested(self._ontology, "ormRoot:ORM2", "orm:ORMModel",
                                           "orm:Objects", "orm:ValueType")
        if orm_value_types:
            for vt in orm_value_types:
                if not vt.get("@IsImplicitBooleanValue"):
                    id = vt["@id"]
                    name = vt["@Name"]
                    data_type = data_types[self._get_nested(vt, "orm:ConceptualDataType", "@ref")]
                    value_types[id] = ORMValueType(id, name, data_type)
        return value_types

    def _parse_data_types(self):
        data_types = {}
        orm_data_types = self._get_nested(self._ontology, "ormRoot:ORM2", "orm:ORMModel", "orm:DataTypes")
        if orm_data_types:
            for k, v in orm_data_types.items():
                if v is not None:
                    data_types[v["@id"]] = k[4:]
        return data_types

    def _parse_roles(self):
        roles = {}
        role_to_player = self._parse_role_player()
        orm_fact_types = self._get_nested(self._ontology, "ormRoot:ORM2", "orm:ORMModel", "orm:Facts", "orm:Fact")
        if orm_fact_types:
            for ft in orm_fact_types:
                relationship_name = (ft['@_Name'])
                orm_roles = self._get_nested(ft, "orm:FactRoles", "orm:Role")
                if orm_roles:
                    roles_list = orm_roles if isinstance(orm_roles, list) else [orm_roles]
                    for ro in roles_list:
                        role_id = ro["@id"]
                        role_name = ro["@Name"]
                        role_player = self._get_nested(ro, "orm:RolePlayer", "@ref")
                        if role_player in self._object_types:
                            roles[role_id] = ORMRole(role_id, role_name, relationship_name, role_to_player[role_id])
        return roles

    def _parse_role_player(self):
        role_to_player = {}
        object_types = ((self._get_nested(self._ontology, "ormRoot:ORM2", "orm:ORMModel", "orm:Objects",
                                          "orm:EntityType") or []) +
                        (self._get_nested(self._ontology, "ormRoot:ORM2", "orm:ORMModel", "orm:Objects",
                                          "orm:ValueType") or []))
        for ot in object_types:
            roles = self._get_nested(ot, "orm:PlayedRoles", "orm:Role")
            if roles:
                roles_list = roles if isinstance(roles, list) else [roles]
                for role in roles_list:
                    role_to_player[role["@ref"]] = ot["@id"]
        return role_to_player

    def _parse_role_value_constraints(self):
        role_value_constraints = {}
        fact_types = self._get_nested(self._ontology, "ormRoot:ORM2", "orm:ORMModel", "orm:Facts", "orm:Fact")
        if fact_types:
            for ft in fact_types:
                roles = self._get_nested(ft, "orm:FactRoles", "orm:Role")
                if roles and isinstance(roles, list):
                    for ro in roles:
                        value_ranges = self._get_nested(ro, "orm:ValueRestriction", "orm:RoleValueConstraint",
                                                        "orm:ValueRanges", "orm:ValueRange")
                        if value_ranges:
                            values = []
                            for rvc in value_ranges:
                                if rvc.get("@MinValue") != rvc.get("@MaxValue"):
                                    raise Exception("Unsupported value range constraint.")
                                values.append(rvc["@MinValue"])
                            role_value_constraints[ro["@id"]] = ORMRoleValueConstraint(self._roles[ro["@id"]], values)
        return role_value_constraints

    def _parse_role_constraints(self):
        constraints = []
        for ec in self._inclusion_constraints.values():
            if set(ec.roles).issubset(self._roles):
                constraints.append(RoleConstraint(ec.roles, ec.exclusive, True))
        for ec in self._exclusion_constraints.values():
            if set(ec.roles).issubset(self._roles):
                constraints.append(RoleConstraint(ec.roles, True, ec.inclusive))
        return constraints

    def _parse_subtype_facts(self):
        subtype_of = defaultdict(list)
        for subtype_arrow in self._parse_subtype_arrows():
            subtype_object = self._build_subtype_object(subtype_arrow)
            subtype_of[subtype_object.supertype_name].append(subtype_object)
        return subtype_of

    def _parse_subtype_arrows(self):
        subtype_of = []
        orm_subtype_facts = self._get_nested(self._ontology, "ormRoot:ORM2", "orm:ORMModel",
                                             "orm:Facts", "orm:SubtypeFact")
        if orm_subtype_facts:
            subtype_facts = orm_subtype_facts if isinstance(orm_subtype_facts, list) else [orm_subtype_facts]
            for sft in subtype_facts:
                subtype_of.append(self._parse_subtype_arrow(sft))
        return subtype_of

    def _parse_subtype_arrow(self, subtype_fact):
        fact_roles = subtype_fact["orm:FactRoles"]
        subtype_id = self._get_nested(fact_roles, "orm:SubtypeMetaRole", "@id")
        subtype = self._get_nested(fact_roles, "orm:SubtypeMetaRole", "orm:RolePlayer", "@ref")
        supertype_id = self._get_nested(fact_roles, "orm:SupertypeMetaRole", "@id")
        supertype = self._get_nested(fact_roles, "orm:SupertypeMetaRole", "orm:RolePlayer", "@ref")
        return ORMSubtypeFact(subtype_id, subtype, supertype_id, supertype)

    def _parse_uniqueness_constraints(self):
        internal = {}
        external = {}
        orm_ucs = self._get_nested(self._ontology, "ormRoot:ORM2", "orm:ORMModel",
                                   "orm:Constraints", "orm:UniquenessConstraint")
        if orm_ucs:
            for uc in orm_ucs:
                uc_id = uc["@id"]
                pid = uc.get("orm:PreferredIdentifierFor", None)
                identifies = None
                if pid is not None and pid["@ref"] in self._entity_types:
                    identifies = pid["@ref"]
                uc_roles = []
                roles = self._get_nested(uc, "orm:RoleSequence", "orm:Role")
                if roles:
                    roles_list = roles if isinstance(roles, list) else [roles]
                    for ro in roles_list:
                        role_id = ro["@ref"]
                        uc_roles.append(role_id)
                constraint = ORMUniquenessConstraint(uc_id, uc_roles, identifies)
                target = internal if uc.get("@IsInternal") == "true" else external
                target[uc_id] = constraint
        return internal, external

    def _process_internal_uniqueness_constraints(self):
        unique_roles = []
        fact_type_to_internal_uc = defaultdict(list)
        fact_type_to_complex_uc = defaultdict(list)
        identifier_fact_type_to_entity_type = defaultdict()

        for uc in self._internal_uniqueness_constraints.values():
            if uc.identifies is None and len(uc.roles) == 1:
                unique_roles.extend(uc.roles)

            role_id = uc.roles[0]
            role = self._roles.get(role_id)
            if not role or not role.relationship_name:
                continue

            fact_type = role.relationship_name

            fact_type_to_internal_uc[fact_type].append(uc)

            if uc.identifies is None:
                if len(uc.roles) > 1:
                    fact_type_to_complex_uc[fact_type].append(uc)
            else:
                identifier_fact_type_to_entity_type[fact_type] = self._entity_types[uc.identifies]

        return unique_roles, fact_type_to_internal_uc, fact_type_to_complex_uc, identifier_fact_type_to_entity_type

    def _parse_mandatory_constraints(self):
        mandatory_constraints = {}
        orm_mcs = self._get_nested(self._ontology, "ormRoot:ORM2", "orm:ORMModel",
                                   "orm:Constraints", "orm:MandatoryConstraint")
        if orm_mcs:
            for mc in orm_mcs:
                mc_id = mc["@id"]
                if mc.get("@IsSimple", None) is not None:
                    mc_roles = []
                    roles = self._get_nested(mc, "orm:RoleSequence", "orm:Role")
                    if roles:
                        roles_list = roles if isinstance(roles, list) else [roles]
                        for ro in roles_list:
                            role_id = ro["@ref"]
                            mc_roles.append(role_id)
                    constraint = ORMMandatoryConstraint(mc_id, mc_roles)
                    mandatory_constraints[mc_id] = constraint
        return mandatory_constraints

    def _parse_inclusion_constraints(self):
        inclusion_constraints = {}
        orm_mcs = self._get_nested(self._ontology, "ormRoot:ORM2", "orm:ORMModel",
                                   "orm:Constraints", "orm:MandatoryConstraint")
        if orm_mcs:
            for mc in orm_mcs:
                mc_id = mc["@id"]
                mc_name = mc["@Name"]
                if mc_name.startswith("InclusiveOrConstraint"):
                    mc_roles = []
                    roles = self._get_nested(mc, "orm:RoleSequence", "orm:Role")
                    if roles:
                        roles_list = roles if isinstance(roles, list) else [roles]
                        for ro in roles_list:
                            role_id = ro["@ref"]
                            mc_roles.append(role_id)
                    exclusive = mc.get("orm:ExclusiveOrExclusionConstraint", None) is not None
                    constraint = ORMInclusionConstraint(mc_id, mc_roles, exclusive)
                    inclusion_constraints[mc_id] = constraint
        return inclusion_constraints

    def _parse_exclusion_constraints(self):
        exclusion_constraints = {}
        orm_mcs = self._get_nested(self._ontology, "ormRoot:ORM2", "orm:ORMModel",
                                   "orm:Constraints", "orm:ExclusionConstraint")
        if orm_mcs:
            for ec in orm_mcs:
                ec_id = ec["@id"]
                ec_roles = []
                role_sequences = self._get_nested(ec, "orm:RoleSequences", "orm:RoleSequence")
                role_sequences_list = role_sequences if isinstance(role_sequences, list) else [role_sequences]
                for rs in role_sequences_list:
                    roles = self._get_nested(rs, "orm:Role")
                    if roles:
                        roles_list = roles if isinstance(roles, list) else [roles]
                        for ro in roles_list:
                            role_id = ro["@ref"]
                            ec_roles.append(role_id)
                inclusive = ec.get("orm:ExclusiveOrMandatoryConstraint", None) is not None
                constraint = ORMExclusionConstraint(ec_id, ec_roles, inclusive)
                exclusion_constraints[ec_id] = constraint
        return exclusion_constraints

    def _parse_ring_constraints(self):
        ring_constraints = {}
        orm_rcs = self._get_nested(self._ontology, "ormRoot:ORM2", "orm:ORMModel",
                                   "orm:Constraints", "orm:RingConstraint")
        if orm_rcs:
            rcs_list = orm_rcs if isinstance(orm_rcs, list) else [orm_rcs]
            for rc in rcs_list:
                rc_id = rc["@id"]
                rc_type = rc["@Type"]
                rc_roles = []
                roles = self._get_nested(rc, "orm:RoleSequence", "orm:Role")
                if roles:
                    roles_list = roles if isinstance(roles, list) else [roles]
                    for ro in roles_list:
                        role_id = ro["@ref"]
                        rc_roles.append(role_id)
                constraint = ORMRingConstraint(rc_id, rc_roles, self.process_ring_constraint(rc_type))
                ring_constraints[rc_id] = constraint
        return ring_constraints

    def _parse_fact_types_reading_orders(self):
        fact_types_readings = defaultdict(list)
        fact_types = self._get_nested(self._ontology, "ormRoot:ORM2", "orm:ORMModel", "orm:Facts", "orm:Fact")
        if fact_types:
            for ft in fact_types:
                relationship_name = (ft['@_Name'])
                reading_orders = self._get_nested(ft, "orm:ReadingOrders", "orm:ReadingOrder")
                if reading_orders:
                    ros = reading_orders if isinstance(reading_orders, list) else [reading_orders]
                    for ro in ros:
                        reading_order = self._parse_fact_types_reading_order(ro)
                        if reading_order:
                            fact_types_readings[relationship_name].append(reading_order)
        return fact_types_readings

    def _parse_fact_types_reading_order(self, reading_order):
        role_texts = self._get_nested(reading_order, "orm:Readings", "orm:Reading", "orm:ExpandedData", "orm:RoleText")
        if not role_texts:
            return None

        role_text_list = role_texts if isinstance(role_texts, list) else [role_texts]

        # Step 1: Build roles with whatever RoleText exists
        index_to_role = {}
        for rt in role_text_list:
            index = int(rt["@RoleIndex"])
            index_to_role[index] = ORMReadingOrderRole(
                index=index,
                prefix=rt.get("@PreBoundText", "").strip() or None,
                postfix=rt.get("@PostBoundText", "").strip() or None,
                text=rt.get("@FollowingText", "").strip() or None,
                role=None
            )

        # Step 2: Ensure all roles from RoleSequence are represented, even if no RoleText exists
        roles = self._get_nested(reading_order, "orm:RoleSequence", "orm:Role")
        if roles:
            roles_list = roles if isinstance(roles, list) else [roles]
            for i, role in enumerate(roles_list):
                role_id = role.get("@ref")
                orm_role = self._roles.get(role_id, None)
                if i in index_to_role:
                    index_to_role[i].role = orm_role
                else:
                    # Fill in with None/defaults if RoleText is missing
                    index_to_role[i] = ORMReadingOrderRole(
                        index=i,
                        prefix=None,
                        postfix=None,
                        text=None,
                        role=orm_role
                    )

        return ORMReadingOrder([index_to_role[i] for i in sorted(index_to_role)])

    @staticmethod
    def _get_nested(d, *keys):
        for key in keys:
            d = d.get(key)
            if d is None:
                return None
        return d

    def _build_subtype_object(self, subtype_arrow):
        sub_name = self._object_types[subtype_arrow.subtype].name
        sup_name = self._object_types[subtype_arrow.supertype].name
        for ec, ec_metadata in self._exclusion_constraints.items():
            if ec_metadata.inclusive and subtype_arrow.supertype_role_id in ec_metadata.roles:
                return ExclusiveInclusiveSubtypeFact(sub_name, sup_name)
            elif subtype_arrow.supertype_role_id in ec_metadata.roles:
                return ExclusiveSubtypeFact(sub_name, sup_name)
        return SubtypeFact(sub_name, sup_name)

    def _parse_mandatory_roles(self):
        mandatory_roles = []
        for mc in self._mandatory_constraints.values():
            mandatory_roles.extend(mc.roles)
        return mandatory_roles

    def _sort_subtype_facts(self):
        # Build a dependency graph to sort subtype facts so that child entity are declared
        # after the parent entity that they extend
        subtype_facts = self._subtype_facts
        nodes = []
        edges = []
        sorted_subtype_facts = []
        # Parent depends on child, i.e., parent entity must be declared before child entity
        for parent in subtype_facts:
            nodes.append(parent)
            for child in subtype_facts[parent]:
                edges.append((parent, child.subtype_name))
        for parent in topological_sort(nodes, edges):
            sorted_subtype_facts.extend(subtype_facts[parent])
        return sorted_subtype_facts

    @staticmethod
    def process_ring_constraint(rc_type):
        pattern = '|'.join(r_type.value for r_type in ORMRingType)
        matches = re.finditer(pattern, rc_type)

        ring_types = []
        for match in matches:
            ring_types.append(ORMRingType(match.group(0)))
        return ring_types



from io import StringIO
from typing import Optional

import pandas as pd

import relationalai.early_access.builder as qb
from relationalai.early_access.builder import define, where
from relationalai.early_access.builder.std.decimals import parse_decimal64, parse_decimal128
from relationalai.early_access.dsl.bindings.common import BindableColumn, AbstractBindableTable
from relationalai.early_access.dsl.snow.common import CsvColumnMetadata
from relationalai.early_access.dsl.utils import normalize
from relationalai.early_access.rel.rel_utils import DECIMAL64_SCALE, DECIMAL128_SCALE


class BindableCsvColumn(BindableColumn):
    _metadata: CsvColumnMetadata
    _column_basic_type: str

    def __init__(self, metadata: CsvColumnMetadata, table: 'CsvTable', model):
        super().__init__(metadata.name, metadata.datatype, table, model)
        self._metadata = metadata
        self._column_basic_type = "Int64" if metadata.datatype._name == qb.Integer._name else "string"

    @property
    def metadata(self):
        return self._metadata

    def basic_type(self):
        return self._column_basic_type

    def decimal_scale(self) -> Optional[int]:
        if self.type() is qb.Decimal64:
            return DECIMAL64_SCALE
        elif self.type() is qb.Decimal128:
            return DECIMAL128_SCALE
        else:
            return None

    def decimal_size(self) -> Optional[int]:
        if self.type() is qb.Decimal64:
            return 64
        elif self.type() is qb.Decimal128:
            return 128
        else:
            return None

    def __repr__(self):
        return f"CSV:{self._source.physical_name()}.{self.physical_name()}"


class CsvTable(AbstractBindableTable[BindableCsvColumn]):
    _basic_type_schema: dict[str, str]
    _csv_data: list[str]

    def __init__(self, name: str, schema: dict[str, qb.Concept], model):
        super().__init__(name, model, set())
        self._initialize(schema, model)

    def _initialize(self, schema: dict[str, qb.Concept], model):
        self._csv_data = list()
        self._cols = {column_name: BindableCsvColumn(CsvColumnMetadata(column_name, column_type), self, model)
                      for column_name, column_type in schema.items()}
        self._basic_type_schema = {col.metadata.name: col.basic_type() for col in self._cols.values()}

    def __str__(self):
        # returns the name of the table, as well as the columns and their types
        return self.physical_name() + ':\n' + '\n'.join(
            [f' {col.metadata.name} {col.metadata.datatype}' for _, col in self._cols.items()]
        ) + '\n' + '\n'.join(
            [f' {fk.source_columns} -> {fk.target_columns}' for fk in self._foreign_keys]
        )

    @property
    def csv_data(self) -> list[str]:
        return self._csv_data

    def physical_name(self) -> str:
        return self._table.lower()

    def data(self, csv_data: str):
        self._csv_data.append(csv_data)
        CsvSourceModule.generate(self, pd.read_csv(StringIO(normalize(csv_data)), dtype=self._basic_type_schema))

class CsvSourceModule:

    @staticmethod
    def generate(table: CsvTable, data: pd.DataFrame):
        for index, row in data.iterrows():
            for column_name in data.columns:
                value = row[column_name]
                if pd.notna(value):
                    column = table.__getattr__(column_name)
                    column_type = column.type()
                    if column_type._name == qb.Date._name:
                        CsvSourceModule._row_to_date_value_rule(column, index, value)
                    elif column_type._name == qb.DateTime._name:
                        CsvSourceModule._row_to_date_time_value_rule(column, index, value)
                    elif column_type._name == qb.Decimal64._name:
                        CsvSourceModule._row_to_decimal64_value_rule(column, index, value)
                    elif column_type._name == qb.Decimal128._name:
                        CsvSourceModule._row_to_decimal128_value_rule(column, index, value)
                    else:
                        CsvSourceModule._row_to_value_rule(column, index, value)

    @staticmethod
    def _row_to_value_rule(column, row, value):
        define(column(row, value))

    @staticmethod
    def _row_to_date_value_rule(column, row, value):
        parse_date = qb.Relationship.builtins['parse_date']
        rez = qb.Date.ref()
        where(parse_date(value, 'Y-m-d', rez)).define(column(row, rez))

    @staticmethod
    def _row_to_date_time_value_rule(column, row, value):
        parse_datetime = qb.Relationship.builtins['parse_datetime']
        rez = qb.DateTime.ref()
        where(parse_datetime(value, 'Y-m-d HH:MM:SS z', rez)).define(column(row, rez))

    @staticmethod
    def _row_to_decimal64_value_rule(column, row, value):
        define(column(row, parse_decimal64(value)))

    @staticmethod
    def _row_to_decimal128_value_rule(column, row, value):
        define(column(row, parse_decimal128(value)))



from __future__ import annotations
from collections import defaultdict
from dataclasses import dataclass
import json
import re
import textwrap
from relationalai.clients.config import Config
from relationalai.clients.util import IdentityParser
from relationalai.early_access.metamodel.util import OrderedSet, ordered_set
from relationalai.early_access.rel.rel_utils import sanitize_identifier
from . import builder as b, annotations as anns
from relationalai import debugging
from snowflake.snowpark.context import get_active_session

#--------------------------------------------------
# Globals
#--------------------------------------------------

_session = None
def get_session():
    global _session
    if not _session:
        try:
            _session = get_active_session()
        except Exception:
            from relationalai.clients.snowflake import Resources
            # TODO: we need a better way to handle global config
            _session = Resources(None, Config(), None).get_sf_session()
    return _session

#--------------------------------------------------
# Snowflake sources
#--------------------------------------------------

SFTypes = {
    "TEXT": "String",
    "FIXED": "Number",
    "DATE": "Date",
    "TIMESTAMP": "DateTime",
    "TIMESTAMP_LTZ": "DateTime",
    "TIMESTAMP_TZ": "DateTime",
    "TIMESTAMP_NTZ": "DateTime",
    "FLOAT": "Float",
    "REAL": "Float",
    "BOOLEAN": "Boolean",
}

SF_ID_REGEX = re.compile(r'^[A-Za-z_][A-Za-z0-9_$]*$')
def quoted(ident:str):
    if SF_ID_REGEX.match(ident) or ident[0] == '"':
        return ident
    return f'"{ident}"'

@dataclass
class TableInfo:
    source:Table|None
    fields:list[b.Field]
    raw_columns:list[dict]

class SchemaInfo:
    def __init__(self, database:str, schema:str) -> None:
        self.database = database
        self.schema = schema
        self.tables = defaultdict(lambda: TableInfo(None, [], []))
        self.fetched = set()

    def fetch(self):
        session = get_session()
        table_names = [name for name in self.tables.keys() if name.upper() not in self.fetched]
        self.fetched.update([x.upper() for x in table_names])
        name_lookup = {x.upper(): x for x in table_names}
        tables = ", ".join([f"'{x.upper()}', '{x}'" for x in self.tables.keys()])
        query = textwrap.dedent(f"""
            begin
                SHOW COLUMNS IN SCHEMA {quoted(self.database)}.{quoted(self.schema)};
                let r resultset := (select "table_name", "column_name", "data_type" from table(result_scan(-1)) as t
                                    where "table_name" in ({tables}));
                return table(r);
            end;
        """)
        with debugging.span("fetch_schema", sql=query):
            columns = session.sql(query).collect()
        for row in columns:
            table_name, column_name, data_type = row
            table_name = name_lookup.get(table_name, table_name)
            info = self.tables[table_name]
            sf_type_info = json.loads(data_type)
            if sf_type_info.get("type") == "FIXED":
                type_str = "Int"
                if sf_type_info.get("scale") > 0:
                    type_str = "Decimal"
            else:
                type_str = SFTypes[sf_type_info.get("type")]
            info.fields.append(b.Field(name=column_name, type_str=type_str, is_many=False))
            info.raw_columns.append(row.as_dict())

class Table():
    _schemas:dict[tuple[str, str], SchemaInfo] = {}
    _used_sources:OrderedSet[Table] = ordered_set()

    def __init__(self, fqn:str, cols:list[str]|None=None) -> None:
        self._fqn = fqn
        parser = IdentityParser(fqn, require_all_parts=True)
        self._database, self._schema, self._table, self._fqn = parser.to_list()
        self._inited = False
        self._ref = b.Integer.ref("row_id")
        self._cols = {}
        self._col_names = cols
        info = self._schemas.get((self._database, self._schema))
        if not info:
            info = self._schemas[(self._database, self._schema)] = SchemaInfo(self._database, self._schema)
        info.tables[self._table].source = self

    def _lazy_init(self):
        if self._inited:
            return
        self._inited = True
        schema_info = self._schemas[(self._database, self._schema)]
        if self._table not in schema_info.fetched:
            schema_info.fetch()
        table_info = schema_info.tables[self._table]
        self._rel = b.Relationship(self._fqn, fields=[b.Field(name="RowId", type_str="RowId")] + table_info.fields)
        self._rel.annotate(anns.external).annotate(anns.from_cdc)

    def __getattr__(self, name: str):
        if name.startswith("_"):
            return super().__getattribute__(name)
        if self._col_names is not None and name not in self._col_names:
            raise ValueError(f"Column {name} not found in {self._fqn}")
        if name.lower() not in self._cols:
            self._cols[name.lower()] = Column(self, name)
        return self._cols[name.lower()]

    def into(self, concept:b.Concept, keys:list[Column]=[]):
        self._lazy_init()
        key_dict = {sanitize_identifier(k._column_name.lower()): k for k in keys}

        # TODO: this is correct, but the rel backend does the wrong thing with it
        # me = concept.new(**key_dict)
        # items = [me]
        # for field in self._rel._fields[1:]:
        #     field_rel = getattr(concept, field.name.lower())
        #     if sanitize_identifier(field.name.lower()) not in key_dict:
        #         items.append(field_rel(me, getattr(self, field.name)))
        # b.where(me).define(*items)

        # this is much less efficient than above

        if keys:
            me = concept.new(**key_dict)
            b.define(me)
        else:
            me = self._rel._field_refs[0]
            b.where(self).define(concept(me))
        for field in self._rel._fields[1:]:
            if sanitize_identifier(field.name.lower()) not in key_dict:
                update = getattr(concept, field.name.lower())(me, getattr(self, field.name))
                b.define(update)

    def _compile_lookup(self, compiler:b.Compiler, ctx:b.CompilerContext):
        self._lazy_init()
        Table._used_sources.add(self)
        compiler.lookup(self._rel, ctx)
        return

class Column(b.Producer):
    def __init__(self, source:Table, column_name:str) -> None:
        super().__init__(None)
        self._column_name = column_name
        self._source = source

    def _compile_lookup(self, compiler:b.Compiler, ctx:b.CompilerContext):
        compiler.lookup(self._source, ctx)
        for ix, field in enumerate(self._source._rel._field_refs):
            assert field._name is not None
            if field._name.lower() == self._column_name.lower():
                return compiler.lookup(b.RelationshipFieldRef(None, self._source._rel, ix), ctx)
        raise ValueError(f"Column {self._column_name} not found in {self._source._fqn}")


from __future__ import annotations

from dataclasses import dataclass
from typing import Optional
from relationalai.early_access.metamodel.rewrite import flatten
from relationalai.early_access.metamodel import ir, factory as f, helpers
from relationalai.early_access.metamodel.compiler import Pass, group_tasks
from relationalai.early_access.metamodel.util import OrderedSet, ordered_set
from relationalai.early_access.metamodel import dependency

class ExtractCommon(Pass):
    """
    Pass to analyze Logical bodies and extract lookups in their own Logical if it makes
    sense. The heuristic is that, if there are multiple lookups, and there are also multiple
    sibling nested logicals that will eventually be extracted by Flatten, then it makes
    sense to extract these logicals into their own "rule", and then make the original body
    just lookup this common rule.

    From:
        Logical
            Logical
                lookup1
                lookup2
                Logical1 ...
                Logical2 ...
    To:
        Logical
            Logical
                lookup1
                lookup2
                derive common
            Logical
                lookup common
                Logical1 ...
                Logical2 ...
    """

    #--------------------------------------------------
    # Public API
    #--------------------------------------------------
    def rewrite(self, model: ir.Model, options:dict={}) -> ir.Model:
        # create the  context
        ctx = ExtractCommon.Context(model)

        # rewrite the root
        replacement = self.handle(model.root, ctx)

        # the new root contains the extracted top level logicals and the rewritten root
        if ctx.rewrite_ctx.top_level:
            new_root = ir.Logical(model.root.engine, tuple(), tuple(ctx.rewrite_ctx.top_level + [replacement]))
        else:
            new_root = replacement

        # create the new model, updating relations and root
        return ir.Model(
            model.engines,
            OrderedSet.from_iterable(model.relations).update(ctx.rewrite_ctx.relations).frozen(),
            model.types,
            new_root
        )

    #--------------------------------------------------
    # IR handlers
    #--------------------------------------------------

    class Context():
        def __init__(self, model: ir.Model):
            self.rewrite_ctx = helpers.RewriteContext()
            self.info = dependency.analyze(model.root)

    def handle(self, task: ir.Task, ctx: Context):
        # currently we only extract if it's a sequence of Logicals, but we could in the
        # future support other intermediate nodes
        if isinstance(task, ir.Logical):
            return self.handle_logical(task, ctx)
        else:
            return task

    def handle_logical(self, task: ir.Logical, ctx: Context):
        # create frame to process the children
        # with ctx_frame(ctx, task) as frame:
            # process the original body
            groups = group_tasks(task.body, {
                "binders": helpers.BINDERS,
                "composites": helpers.COMPOSITES
            })
            binders = groups["binders"]
            composites = groups["composites"]

            # the new body of the rewritten task
            body:OrderedSet[ir.Task] = ordered_set()

            # quick check to see if it's worth doing more analysis; we only want to extract
            # common binders if there are multiple, and there are also multiple composites
            # that will be extracted by the flatten pass later (so that they can share the
            # extracted logic).
            plan = None
            if len(binders) > 1 and composites:
                extractables = flatten.extractables(composites)
                # only makes sense to extract common if at least one nested composite will be extracted
                if extractables:
                    # make a plan to extract common tasks from the logical
                    plan = self._create_extraction_plan(binders, composites, extractables, ctx)
                    if plan and len(composites) > 1:
                        # plan is worthwhile and there are multiple composites, extract the common body and add the connection to the body
                        exposed_vars = plan.exposed_vars.get_list()
                        plan.common_reference = f.lookup(helpers.extract(task, plan.common_body, exposed_vars, ctx.rewrite_ctx, "common"), exposed_vars)
                        # if we are not distributing the reference, add to the main body
                        if not plan.distribute_common_reference:
                            body.add(plan.common_reference)

            # if we have a plan and will distribute the common reference, keep track of
            # variables still needed by the remaining tasks, as they need to be hoisted by
            # the remaining composites that get the common reference
            remaining_vars = None
            if plan and plan.distribute_common_reference:
                # add variables hoisted by this logical that are in the exposed vars, to
                # make sure they are hoisted all the way through
                remaining_vars = OrderedSet.from_iterable(helpers.hoisted_vars(task.hoisted)) & plan.exposed_vars
                for child in task.body:
                    if child in groups["other"] or child not in plan.remaining_body or child in composites:
                        continue
                    remaining_vars.update(ctx.info.task_inputs(child))
                    remaining_vars.update(ctx.info.task_outputs(child))
                remaining_vars = remaining_vars & plan.exposed_vars

            # if the plan was not used in one of the cases above, ignore it completely, we
            # are not extracting common nor distributing it around
            if plan and not plan.distribute_common_reference and not len(composites) > 1:
                plan = None

            # recursively handle children
            for child in task.body:
                # skip children that were extracted
                if plan and child not in groups["other"] and child not in plan.remaining_body and child not in composites:
                    continue

                # no plan or child is not a composite, so just add the handled to the body
                if not plan or child not in composites:
                    body.add(self.handle(child, ctx))
                    continue

                # there is a plan and the child is in composites, so...
                replacement = self.handle(child, ctx)

                # this child needs either extra local dependencies or the common reference
                if child in plan.local_dependencies or plan.distribute_common_reference:
                    # the new body will have maybe the common reference and the local deps
                    replacement_body = ordered_set()
                    hoisted = OrderedSet.from_iterable(replacement.hoisted if isinstance(replacement, helpers.COMPOSITES) else [])
                    if plan.distribute_common_reference:
                        if len(composites) == 1:
                            # if there's a single composite, just insert the whole common body into it
                            replacement_body.update(plan.common_body)
                        else:
                            # otherwise insert a clone of the reference on the extracted rule
                            assert(plan.common_reference)
                            replacement_body.add(plan.common_reference.clone())
                        # add remaining vars to hoisted, making sure there's no duplicates (due to VarOrDefault)
                        hoisted_vars = helpers.hoisted_vars(hoisted)
                        if remaining_vars:
                            hoisted = OrderedSet.from_iterable(filter(lambda v: v not in hoisted_vars, remaining_vars)) | hoisted

                    if child in plan.local_dependencies:
                        for local_dep in plan.local_dependencies[child]:
                            replacement_body.add(local_dep.clone())
                        #replacement_body.update(plan.local_dependencies[child])

                    if isinstance(replacement, ir.Logical):
                        # if the replacements is a logical, we can just add to the body
                        body.add(replacement.reconstruct(
                            replacement.engine,
                            tuple(hoisted.get_list()),
                            tuple(replacement_body.update(replacement.body).get_list()),
                            replacement.annotations
                        ))
                    else:
                        # anything else is now wrapped in a logical so that we can add the
                        # local dependencies
                        body.add(f.logical(replacement_body.add(replacement).get_list(), hoisted.get_list(), replacement.engine))
                else:
                    # child does not need extras in the body, just add it to the main body
                    body.add(replacement)

            return ir.Logical(task.engine, task.hoisted, tuple(body))

    @dataclass
    class ExtractionPlan():
        # tasks to extract to the body of the common logical
        common_body: OrderedSet[ir.Task]
        # tasks to remain in the original body
        remaining_body: OrderedSet[ir.Task]
        # variables to be exposed by the common logical
        exposed_vars: OrderedSet[ir.Var]
        # map from nested composite to the tasks in the common body that still need to be
        # included in its body, because it contains variables not exposed by the common logical
        local_dependencies: dict[ir.Task, OrderedSet[ir.Task]]
        # whether the common reference should be distributed to composites
        distribute_common_reference: bool
        # a reference to the common connection created for this plan, if any
        common_reference: Optional[ir.Lookup] = None


    def _create_extraction_plan(self, binders: OrderedSet[ir.Task], composites: OrderedSet[ir.Task], extractables: list[ir.Task], ctx: Context):
        """
        Compute a plan to extract tasks in this frame that are common dependencies
        across these composite tasks.
        """
        # compute intersection of task dependencies and inputs
        sample = composites.some()
        deps = ctx.info.task_dependencies(sample)
        if deps is None:
            return None
        # only get sibling dependencies
        common_body = binders & deps
        exposed_vars = OrderedSet.from_iterable(ctx.info.task_inputs(sample))

        # now extract from the original sets
        for composite in composites:
            if composite is sample:
                continue

            # compute sibling dependencies
            deps = ctx.info.task_dependencies(composite)
            if deps:
                for task in common_body:
                    if task not in deps:
                        common_body.remove(task)
            # compute common input vars
            t_inputs = ctx.info.task_inputs(composite)
            for v in exposed_vars:
                if v not in t_inputs:
                    exposed_vars.remove(v)

        # no vars in common, not worth to extract
        if not exposed_vars:
            return None

        for task in common_body:
            local_deps = ctx.info.local_dependencies(task)
            if local_deps:
                common_body.update(local_deps & binders)

        # not useful to extract common tasks if there's a single one
        if len(common_body) < 2:
            return None

        # check if some variable used in the common body is needed by some binder that is
        # not going to be extracted. In that case, we need to expose this variable from the
        # common body
        common_vars = ordered_set()
        for task in common_body:
            common_vars.update(ctx.info.task_outputs(task))
        common_vars = common_vars - exposed_vars
        for v in common_vars:
            for binder in binders:
                if binder not in common_body and ctx.info.task_inputs(binder) and v in ctx.info.task_inputs(binder):
                    exposed_vars.add(v)
                    break

        # check which of the original binders remain, and make sure their dependencies also stay
        remaining = ordered_set()
        for binder in binders:
            if binder not in common_body:
                remaining.add(binder)
                deps = self._compute_local_dependencies(ctx, binders, binder, common_body, exposed_vars)
                if deps:
                    remaining.update(deps)

        # for each composite, check if there are additional tasks needed, because the task
        # depends on it but it is not exposed by the vars
        local_dependencies: dict[ir.Task, OrderedSet[ir.Task]] = dict()
        for composite in composites:
            local = self._compute_local_dependencies(ctx, binders, composite, common_body, exposed_vars)
            if local:
                local_dependencies[composite] = local

        # distribute the common reference only if all of the composites are extractable and there's nothing else remaining
        distribute_common_reference = len(extractables) == len(composites) and not remaining
        return ExtractCommon.ExtractionPlan(common_body, remaining, exposed_vars, local_dependencies, distribute_common_reference)


    def _compute_local_dependencies(self, ctx: Context, binders: OrderedSet[ir.Task], composite: ir.Task, common_body: OrderedSet[ir.Task], exposed_vars: OrderedSet[ir.Var]):
        """
        The tasks in common_body will be extracted into a logical that will expose the exposed_vars.
        Compute which additional dependencies are needed specifically for this composite, because
        it depends on some tasks that are extracted to common_body but not exposed by exposed_vars.
        """

        # working list of vars we still need to fulfill
        inputs = ctx.info.task_inputs(composite)
        if not inputs:
            return None

        # vars exposed by exposed vars + tasks added to the local body
        vars_exposed = OrderedSet.from_iterable(exposed_vars)
        vars_needed = (inputs - vars_exposed)
        if not vars_needed:
            return None

        # needs = ctx.info.local_dependencies(composite) & binders | common_body
        # if needs:
        #     return needs
        # return None

        # this is a greedy algorithm that uses the first task in the common body that provides
        # a variable needed; it may result in sub-optimal extraction, but should be correct
        local_body = ordered_set()
        while(vars_needed):
            v = vars_needed.pop()
            for x in common_body:
                if x not in local_body:
                    # an x that is not yet in local_body can fulfill v
                    x_outputs = ctx.info.task_outputs(x)
                    if x_outputs and v in x_outputs:
                        # add it to local_body and add its outputs to vars exposed
                        local_body.add(x)
                        vars_exposed.add(x_outputs)
                        # but add its inputs the vars now needed
                        inputs = ctx.info.task_inputs(x)
                        if inputs:
                            vars_needed.update(inputs - vars_exposed)
        return local_body





    def _create_extraction_plan_old(self, binders: OrderedSet[ir.Task], composites: OrderedSet[ir.Task], extractables: list[ir.Task], ctx: Context):
        """
        Compute a plan to extract tasks in this frame that are common dependencies
        across these composite tasks.
        """
        # compute intersection of task dependencies and inputs
        sample = composites.some()
        deps = ctx.info.task_dependencies(sample)
        if deps is None:
            return None
        # only get sibling dependencies
        common_body = binders & deps
        exposed_vars = OrderedSet.from_iterable(ctx.info.task_inputs(sample))

        # now extract from the original sets
        for composite in composites:
            if composite is sample:
                continue

            # compute sibling dependencies
            deps = ctx.info.task_dependencies(composite)
            if deps:
                for task in common_body:
                    if task not in deps:
                        common_body.remove(task)
            # compute common input vars
            t_inputs = ctx.info.task_inputs(composite)
            for v in exposed_vars:
                if v not in t_inputs:
                    exposed_vars.remove(v)

        # no vars in common, not worth to extract
        if not exposed_vars:
            return None

        # pull the transitive closure of the intersected tasks in the common body
        chasing = True
        while(chasing):
            chasing = False
            for task in common_body:
                deps = ctx.info.task_dependencies(task)
                if deps:
                    for hop in binders & deps:
                        if hop not in common_body:
                            common_body.add(hop)
                            chasing = True

        # not useful to extract common tasks if there's a single one
        if len(common_body) < 2:
            return None

        # check if some variable used in the common body is needed by some binder that is
        # not going to be extracted. In that case, we need to expose this variable from the
        # common body
        common_vars = ordered_set()
        for task in common_body:
            common_vars.update(ctx.info.task_outputs(task))
        common_vars = common_vars - exposed_vars
        for v in common_vars:
            for binder in binders:
                if binder not in common_body and ctx.info.task_inputs(binder) and v in ctx.info.task_inputs(binder):
                    exposed_vars.add(v)
                    break

        # check with of the original binders remain, and make sure their dependencies also stay
        remaining = ordered_set()
        for binder in binders:
            if binder not in common_body:
                remaining.add(binder)
                deps = self._compute_local_dependencies_old(ctx, binder, common_body, exposed_vars)
                if deps:
                    remaining.update(deps)

        # for each composite, check if there are additional tasks needed, because the task
        # depends on it but it is not exposed by the vars
        local_dependencies: dict[ir.Task, OrderedSet[ir.Task]] = dict()
        for composite in composites:
            local = self._compute_local_dependencies_old(ctx, composite, common_body, exposed_vars)
            if local:
                local_dependencies[composite] = local

        # distribute the common reference only if all of the composites are extractable and there's nothing else remaining
        distribute_common_reference = len(extractables) == len(composites) and not remaining
        # TODO: we are forcing this for now, but we should remove the whole analysis
        distribute_common_reference = False
        return ExtractCommon.ExtractionPlan(common_body, remaining, exposed_vars, local_dependencies, distribute_common_reference)


    def _compute_local_dependencies_old(self, ctx: Context, composite: ir.Task, common_body: OrderedSet[ir.Task], exposed_vars: OrderedSet[ir.Var]):
        """
        The tasks in common_body will be extracted into a logical that will expose the exposed_vars.
        Compute which additional dependencies are needed specifically to this composite, because
        it depends on some tasks that are extracted to common_body but not exposed by exposed_vars.
        """

        # vars exposed by exposed vars + tasks added to the local body
        vars_exposed = OrderedSet.from_iterable(exposed_vars)
        # working list of vars we still need to fulfill
        inputs = ctx.info.task_inputs(composite)
        if not inputs:
            return None
        vars_needed = (inputs - vars_exposed)
        if not vars_needed:
            return None

        # this is a greedy algorithm that uses the first task in the common body that provides
        # a variable needed; it may result in sub-optimal extraction, but should be correct
        local_body = ordered_set()
        while(vars_needed):
            v = vars_needed.pop()
            for x in common_body:
                if x not in local_body:
                    # an x that is not yet in local_body can fulfill v
                    x_outputs = ctx.info.task_outputs(x)
                    if x_outputs and v in x_outputs:
                        # add it to local_body and add its outputs to vars exposed
                        local_body.add(x)
                        vars_exposed.add(x_outputs)
                        # but add its inputs the vars now needed
                        inputs = ctx.info.task_inputs(x)
                        if inputs:
                            vars_needed.update(inputs - vars_exposed)
        return local_body



from __future__ import annotations

import secrets
import string

import relationalai as rai

from pandas import DataFrame
from typing import Any, Union, cast, Optional

from .. import Compiler
from relationalai.early_access.metamodel.util import ordered_set
from relationalai.early_access.metamodel import ir, executor as e, factory as f

class SnowflakeExecutor(e.Executor):
    """Executes SQL using the RAI client."""

    def __init__(
            self,
            database: str,
            schema: str,
            dry_run: bool = False,
            skip_denormalization: bool = False,
            config: rai.Config | None = None,
    ) -> None:
        super().__init__()
        self.database = database
        self.schema = schema
        self.dry_run = dry_run
        self.config = config or rai.Config()
        self.compiler = Compiler(skip_denormalization)
        self.provider = cast(rai.clients.snowflake.Provider, rai.Provider(config=config))

    def execute(self, model: ir.Model, task:ir.Task, result_cols:Optional[list[str]]=None,
                export_to:Optional[str]=None, update:bool=False) -> Union[DataFrame, Any]:
        """ Execute the SQL query directly. """
        if self.dry_run:
            return DataFrame()

        db_name = f"{self.database}_{self._generate_unique_id()}"
        db_query = f"CREATE OR REPLACE DATABASE {db_name};"
        schema_query = f"CREATE OR REPLACE SCHEMA {db_name}.{self.schema};"
        use_schema_query = f"USE SCHEMA {db_name}.{self.schema};"
        model_sql = self.compiler.compile(model)

        full_sql_model = f"{db_query}\n{schema_query}\n{use_schema_query}\n{model_sql}"

        # TODO: find the way how to compile task instead of building model with one root task
        query_model = f.model(ordered_set(), ordered_set(), ordered_set(), task)
        query_sql = self.compiler.compile(query_model)

        try:
            self.provider.resources._session.connection.execute_string(full_sql_model) # type: ignore
            result = cast(DataFrame, self.provider.sql(query_sql, format='pandas'))
            return result
        finally:
            self.provider.sql(f"DROP DATABASE IF EXISTS {db_name};")

    def _generate_unique_id(self, length=7):
        alphabet = string.ascii_letters + string.digits  # a-zA-Z0-9
        return ''.join(secrets.choice(alphabet) for _ in range(length))


from abc import abstractmethod
from typing import Any

from relationalai.early_access.dsl.core.cset import UnstableConstraintSet
from relationalai.early_access.dsl.core.exprs.relational import RelationVariable
from relationalai.early_access.dsl.core.exprs.scalar import ScalarVariable

# We use TypeVariables to formalize the meaning of a ConstrainedValueType.
# Each represents what would be a variable in a ValueType or ValueSubtype
# declaration.
#
class TypeVariable:

    # Each TypeVariable declares some AbstractValueType (vt) which might
    # be a ValueSubtype. In addition, each TypeVariable has a Position
    # that is used to form the name of the variable when needed.
    #
    def __init__(self, vt, vmap={}):
        self.pos = None
        self._cset = UnstableConstraintSet()
        self.decl = vt
        self._vmap = vmap

    # Returns the set of ScalarConstraints that this TypeVariable denotes
    #
    def denotes(self, C={}):
        for sc in self._cset._sconstraints.values():
            C[sc.entityid()] = sc

        for ac in self._cset._atoms.values():
            # [REKS] Check the following logic carefully to make sure we don't accidentally
            #        capture any variable names duirng interpretation of type atoms
            #
            scs = self.decl.interpret_type_atom(ac)
            for s in scs:
                if s not in C:
                    C[s] = scs[s]
        return C

    @abstractmethod
    def dupe(self, vmap) -> tuple['TypeVariable', Any]: pass

    def setpos(self, pos): pass

    @abstractmethod
    def varname(self) -> str: pass

class BasicTypeVariable(ScalarVariable, TypeVariable):

    # An UnconstrainedTypeVarible ranges over some UnconstrainedValueType,
    # which will be the root unconstrained type of vt.
    #
    def __init__(self, vt, vmap={}):
        if vt.nominal():
            raise Exception(f"Cannot create BasicTypeVariable of nominal type {vt._name}")
        super().__init__(vt, vmap)

    def display(self):
        return self.varname()

    def dupe(self, vmap={}):
        d = BasicTypeVariable(self.decl, vmap)
        vmap[id(self)] = d
        cs = self._cset._sconstraints
        for c in cs.values():
            newc = c.revar(vmap)
            d._cset._sconstraints[id(newc)] = newc
        return (d, vmap)

    def entityid(self):
        return hash(self.varname())

    def grounded(self):
        return False

    def grounded_using(self, groundings):
        return False

    def physical_typeof(self):
        return self.decl.root_unconstrained_type()

    def rename(self, renaming):
        nm = self.varname()
        if nm in renaming:
            return renaming[nm]
        else:
            return self

    def revar(self, vmap):
        nm = id(self)
        if nm in vmap:
            return vmap[nm]
        else:
            return self

    def setpos(self, pos):
        self.pos = pos
        self._cset = self._cset.stabilize()
        return pos+1

    def to_z3(self, vars):
        return vars[self.display()]

    def typeof(self):
        return self.decl

    def typevar_tuple(self):
        return self

    def varname(self):
        return f"x{str(self.pos)}"

class NominalTypeVariable(RelationVariable, TypeVariable):

    # If vt is nominal, then the TypeVariable itself is nominal
    # and ranges over some root nominal ValueType.
    #
    def __init__(self, vt, vmap={}):
        super().__init__(vt, vmap)
        self._parts = []

    def all_constraint_display(self):
        parts = [self.constraint_display()]
        for p in self._parts:
            if isinstance(p, NominalTypeVariable):
                parts.append(p.all_constraint_display())
        return "\n".join(parts)

    def constraint_display(self):
        parts = [ p.varname() for p in self._parts ]
        args = ", ".join(parts)
        return f"   {self.varname()}({args})"

    def denotes(self, C={}):
        C = super().denotes(C)
        for p in self._parts:
            C = p.denotes(C)
        return C

    def display(self):
        return self.varname()

    def dupe(self, vmap={}):
        d = NominalTypeVariable(self.decl, vmap)
        vmap[id(self)] = d

        for p in self._parts:
            (pprime, vmap) = p.dupe(vmap)
            d._parts.append(pprime)

        vmap = d._vmap

        cs = self._cset._sconstraints
        for c in cs.values():
            newc = c.revar(vmap)
            d._cset._sconstraints[id(newc)] = newc
        acs = self._cset._atoms
        for ac in acs.values():
            newac = ac.revar(vmap)
            d._cset._atoms[id(newac)] = newac
        return (d, vmap)

    def entityid(self):
        return hash(self.varname())

    def rename(self, renaming):
        nm = self.varname()
        if nm in renaming:
            return renaming[nm]
        else:
            return self

    def revar(self, vmap):
        nm = id(self)
        if nm in vmap:
            return vmap[nm]
        else:
            return self

    def setpos(self, pos):
        self.pos = pos
        pos = pos + 1
        for p in self._parts:
            pos = p.setpos(pos)
        self._cset = self._cset.stabilize()
        return pos

    def typeof(self):
        return self.decl

    def typevar_tuple(self):
        tvars = []

        for tvar in self._parts:
            if isinstance(tvar, NominalTypeVariable):
                tvars.append(tvar.typevar_tuple())
            else:
                tvars.append(tvar)
        if len(tvars) == 1:
            result = tvars[0]
        else:
            result = tuple(tvars)
        return result

    def varname(self):
        return f"y{str(self.pos)}"



from __future__ import annotations
import time
from typing import Any, List
from dataclasses import dataclass
import textwrap
from .. import dsl, std
from ..std import rel
from ..metamodel import Builtins
from ..tools.cli_controls import Spinner
from .. import debugging
import uuid
import relationalai
import json
from ..clients.util import poll_with_specified_overhead
from ..clients.snowflake import Resources as SnowflakeResources

rel_sv = rel._tagged(Builtins.SingleValued)

APP_NAME = relationalai.clients.snowflake.APP_NAME

ENGINE_TYPE_SOLVER = "SOLVER"

# --------------------------------------------------
# SolverModel object.
# --------------------------------------------------

class SolverModel:
    def __init__(self, graph:dsl.Graph):
        self.graph = graph
        self.id = dsl.next_id()
        self.scope = f"solvermodel{self.id}_"
        scope = self.scope
        self.Variable = dsl.Type(graph, "variables", scope=scope)
        self.MinObjective = dsl.Type(graph, "min_objectives", scope=scope)
        self.MaxObjective = dsl.Type(graph, "max_objectives", scope=scope)
        self.Constraint = dsl.Type(graph, "constraints", scope=scope)
        self.Solution = dsl.Type(graph, "solutions", scope=scope)
        self.components = [
            (self.MinObjective, "minimization objectives"),
            (self.MaxObjective, "maximization objectives"),
            (self.Constraint, "constraints"),
        ]
        self.solve_index = 0
        self.is_solved = False

        # Install model helpers.
        self.graph.install_raw(textwrap.dedent(f"""
            @inline
            def _solverlib_ho_appl(op, {{R}}, s): rel_primitive_solverlib_ho_appl(R, op, s)

            @inline
            def _solver_unwrap({{R}}, h, x...): exists((v) | R(v, x...) and pyrel_unwrap(v, h))

            declare {scope}variable_name
            declare {scope}component_name
            declare {scope}serialized
            declare {scope}primal_start

            def {scope}component_string(h, s):
                rel_primitive_solverlib_print_expr(
                    {scope}serialized[h], _solver_unwrap[{scope}variable_name], s
                )

            def {scope}specialized_components(t, h, s):
                exists((v) | {{
                    (:min_objective, {scope}min_objectives);
                    (:max_objective, {scope}max_objectives);
                    (:constraint, {scope}constraints);
                    }}(t, v) and pyrel_unwrap(v, h) and {scope}serialized(v, s)
                )

            declare {scope}solve_output
        """))
        return None

    # Add an entity to the variable set, (optionally) set a string name from the
    # arguments and add domain constraints on the variable.
    def variable(
        self,
        var, # variable entity
        name_args:List|None=None, # list of strings to concatenate into a string name
        type:str|None=None, # variable type: "integer" or "zero_one"
        lower:int|float|None=None, # lower bound
        upper:int|float|None=None, # upper bound
        fixed:int|float|None=None, # fixed value
        start:int|float|None=None, # (primal) start value
    ):
        if type not in {"integer", "zero_one", None}:
            raise Exception(f"Invalid domain type: {type}.")
        var.set(self.Variable)

        # Set variable name.
        if name_args:
            var.set(**{f"{self.scope}variable_name": make_string(name_args)})

        # Add domain constraints.
        cons = []
        if fixed is not None:
            cons.append(eq(var, fixed))
        if type == "zero_one":
            cons.append(zero_one(var))
        if lower is not None and upper is not None:
            if type == "integer":
                cons.append(integer_interval(var, lower, upper))
            else:
                cons.append(interval(var, lower, upper))
        else:
            if type == "integer":
                cons.append(integer(var))
            if lower is not None:
                cons.append(gte(var, lower))
            if upper is not None:
                cons.append(lte(var, upper))
        if len(cons) == 1:
            self.constraint(cons[0])
        elif len(cons) > 1:
            self.constraint(and_(*cons))

        # Set primal start.
        if start is not None:
            var.set(**{f"{self.scope}primal_start": start})
        return var

    # Get variable string name.
    def variable_name(self, var):
        return std.alias(getattr(var, f"{self.scope}variable_name"), "name")

    # Add a constraint, minimization objective, or maximization objective.
    def constraint(self, expr, name_args:List|None=None):
        return self._add_component(self.Constraint, expr, name_args)

    def min_objective(self, expr, name_args:List|None=None):
        return self._add_component(self.MinObjective, expr, name_args)

    def max_objective(self, expr, name_args:List|None=None):
        return self._add_component(self.MaxObjective, expr, name_args)

    def _add_component(self, typ, expr, name_args:List|None):
        # TODO(coey) for now we create unique property names for the serialized component
        # property in order to get unique relation names in the generated Rel, otherwise
        # we have trouble tracing performance data back to the original PyRel block.
        prop = "serialized" + str(dsl.next_id())
        comp = typ.add(**{prop: _wrap_expr(expr)})
        self.graph.install_raw(f"def {self.scope}serialized {{ {self.scope}{prop} }}")
        # comp = typ.add(serialized=_wrap_expr(expr)) # TODO(coey) the simpler way.
        if name_args:
            comp.set(component_name=make_string(name_args))
        return comp

    # Get component string name.
    def component_name(self, comp):
        return std.alias(comp.component_name, "name")

    # Get serialized component string in human-readable format.
    def component_string(self, comp):
        return std.alias(comp.component_string, "string")

    # Summarize the model by printing the number of variables and components.
    # Use outside a rule/query.
    def summarize(self):
        with self.graph.query() as select:
            vars = select(std.aggregates.count(self.Variable()))
        s = f"Model has: {vars.results.iat[0, 0]} variables"
        for (c_type, c_name) in self.components:
            with self.graph.query() as select:
                exprs = select(std.aggregates.count(c_type()))
            if not exprs.results.empty:
                s += f", {exprs.results.iat[0, 0]} {c_name}"
        print(s)
        return None

    # Print the model in human-readable format. Use outside a rule/query.
    def print(self):
        with self.graph.query() as select:
            vars = select(rel.last(getattr(rel, f"{self.scope}variable_name")))
        print("variables:")
        print(vars.results.to_string(index=False, header=False))
        for (c_type, c_name) in self.components:
            with self.graph.query() as select:
                exprs = select(self.component_string(c_type()))
            if not exprs.results.empty:
                print(c_name + ":")
                print(exprs.results.to_string(index=False, header=False))
        return None

    # Solve the model given a solver and solver options. Use outside a rule/query.
    def solve(self, solver: Solver, log_to_console=True, **kwargs):
        self.is_solved = False
        self.solve_index += 1
        self.solve_output = dsl.RelationNS([f"{self.scope}solve_output"], f"i_{self.solve_index}")

        options = kwargs
        options["version"] = 1

        # Validate options.
        for k, v in options.items():
            if not isinstance(k, str):
                raise Exception(f"Invalid parameter key. Expected string, got {type(k)} for {k}.")
            if not isinstance(v, (int, float, str, bool)):
                raise Exception(
                    f"Invalid parameter value. Expected string, integer, float, or boolean, got {type(v)} for {k}."
                )

        # Run the solve query and insert the solve_output result.
        scope = self.scope
        variable_name_string = f"{scope}variable_name" if "print_format" in options else "{}"
        component_name_string = f"{scope}component_name" if "print_format" in options else "{}"

        input_id = uuid.uuid4()
        model_uri = f"snowflake://APP_STATE.RAI_INTERNAL_STAGE/job-inputs/solver/{input_id}/model.binpb"
        sf_input_uri = f"snowflake://job-inputs/solver/{input_id}/model.binpb"

        payload: dict[str, Any] = {"solver": solver.solver_name.lower()}
        payload["options"] = options
        payload["model_uri"] = sf_input_uri

        # 1. Materialize the model and store it.
        # TODO(coey) Currently we must run a dummy query to install the pyrel rules in a separate txn
        # to the solve_output updates. Ideally pyrel would offer an option to flush the rules separately.
        self.graph.exec_raw("")
        self.graph.exec_raw(
            textwrap.dedent(f"""
            ic model_not_empty("Solver model is empty.") requires not empty({scope}model_string)

            @no_diagnostics(:EXPERIMENTAL)
            def {scope}model_string {{
                rel_primitive_solverlib_model_string[{{
                    (:variable, _solver_unwrap[{scope}variables]);
                    {scope}specialized_components;
                    (:variable_name, _solver_unwrap[{variable_name_string}]);
                    (:expression_name, _solver_unwrap[{component_name_string}]);
                    (:primal_start, _solver_unwrap[{scope}primal_start]);
                }}]
            }}

            def config[:envelope, :content_type]: "application/octet-stream"
            def config[:envelope, :payload, :data]: {scope}model_string
            def config[:envelope, :payload, :path]: "{model_uri}"
            def export {{ config }}
            """)
        )

        # 2. Execute job and wait for completion.
        job_id = solver._exec_job(payload, log_to_console=log_to_console)

        # 3. Extract result.
        res = self.graph.exec_raw(
            textwrap.dedent(f"""
            ic result_not_empty("Solver result is empty.") requires not empty(result)

            def result {{load_binary["snowflake://APP_STATE.RAI_INTERNAL_STAGE/job-results/{job_id}/result.binpb"] }}
            def delete[:{scope}solve_output, :"i_{self.solve_index}"]: {scope}solve_output[:"i_{self.solve_index}"]

            @no_diagnostics(:EXPERIMENTAL)
            def insert[:{scope}solve_output, :"i_{self.solve_index}"]:
                rel_primitive_solverlib_extract[result]

            def output[:solver_error]: {scope}solve_output[:"i_{self.solve_index}", :error]
            """),
            readonly=False,
        )
        errors = []
        for result in res.results:
            if result["relationId"] == "/:output/:solver_error/String":
                errors.extend(result["table"]["v1"])

        # 4. Map results to solution.
        with self.graph.rule(dynamic=True):
            sol = self.Solution.add(index = self.solve_index)
            for name in {"error", "termination_status", "solve_time_sec", "objective_value", "solver_version", "printed_model"}:
                val = dsl.create_var()
                getattr(self.solve_output, name)(val)
                sol.set(**{name:val})

        self.is_solved = True

        if len(errors) > 0:
            raise Exception("\n".join(errors))

        return None

    # Get scalar result information after solving.
    def __getattr__(self, name:str):
        if not self.is_solved:
            raise Exception("Model has not been solved yet.")
        if name in {"error", "termination_status", "solve_time_sec", "objective_value", "solver_version", "printed_model"}:
            return getattr(self.Solution(index = self.solve_index), name)
        else:
            return None

    # Get variable point values after solving. If `index` is specified, get the value
    # of the variable in the return the `index`-th solution.
    def value(self, var, index:int|None=None):
        if not self.is_solved:
            raise Exception("Model has not been solved yet.")
        val = dsl.create_var()
        unwrap_var = rel_sv.pyrel_unwrap(var)
        if index:
            self.solve_output.points(unwrap_var, index, val)
        else:
            self.solve_output.point(unwrap_var, val)
        std.alias(val, "value")
        return val

# --------------------------------------------------
# Operator definitions
# --------------------------------------------------

# Builtin binary operators

def plus(left, right):
    return _make_fo_expr(10, left, right)

def minus(left, right):
    return _make_fo_expr(11, left, right)

def mult(left, right):
    return _make_fo_expr(12, left, right)

def div(left, right):
    return _make_fo_expr(13, left, right)

def pow(left, right):
    return _make_fo_expr(14, left, right)

def eq(left, right):
    return _make_fo_expr(30, left, right)

def neq(left, right):
    return _make_fo_expr(31, left, right)

def lte(left, right):
    return _make_fo_expr(32, left, right)

def gte(left, right):
    return _make_fo_expr(33, left, right)

def lt(left, right):
    return _make_fo_expr(34, left, right)

def gt(left, right):
    return _make_fo_expr(35, left, right)

# First order operators

def abs(arg):
    return _make_fo_expr(20, arg)

def exp(arg):
    return _make_fo_expr(21, arg)

def log(arg):
    return _make_fo_expr(22, arg)

def integer(arg):
    return _make_fo_expr(41, arg)

def zero_one(arg):
    return _make_fo_expr(42, arg)

def interval(arg, low, high):
    return _make_fo_expr(51, low, high, arg)

def integer_interval(arg, low, high):
    return _make_fo_expr(50, low, high, 1, arg)

def if_then_else(cond, left, right):
    return _make_fo_expr(60, cond, left, right)

def not_(arg):
    return _make_fo_expr(61, arg)

def implies(left, right):
    return _make_fo_expr(62, left, right)

def iff(left, right):
    return _make_fo_expr(63, left, right)

def xor(left, right):
    return _make_fo_expr(64, left, right)

def and_(*args):
    return _make_fo_expr(70, *args)

def or_(*args):
    return _make_fo_expr(71, *args)

# Aggregate operators

def sum(*args, per=[]) -> Any:
    return _make_ho_expr(80, args, per)

def product(*args, per=[]) -> Any:
    return _make_ho_expr(81, args, per)

def min(*args, per=[]) -> Any:
    return _make_ho_expr(82, args, per)

def max(*args, per=[]) -> Any:
    return _make_ho_expr(83, args, per)

def count(*args, per=[]) -> Any:
    return _make_ho_expr(84, args, per)

def all_different(*args, per=[]) -> Any:
    return _make_ho_expr(90, args, per)

# --------------------------------------------------
# Symbolic expression helpers
# --------------------------------------------------

def _make_fo_expr(*args):
    expr = rel_sv.rel_primitive_solverlib_fo_appl(*args)
    expr.__class__ = SolverExpression
    return expr

# TODO(coey) test:
# dsl.tag(rel_sv.rel_primitive_solverlib_fo_appl, Builtins.Expensive)

def _make_ho_expr(op, args, per):
    return SolverExpression(dsl.get_graph(), _ho_appl_def, [args, per, [op]])

_ho_appl_def = dsl.build.aggregate_def("_solverlib_ho_appl")

class SolverExpression(dsl.Expression):
    def __init__(self, graph, op, args):
        super().__init__(graph, op, args)

def _wrap_expr(e):
    # If expression is not known to produce a serialized expression string,
    # wrap it with the identity operation just in case
    return e if isinstance(e, SolverExpression) else _make_fo_expr(0, e)

# Symbolic expression context, in which some builtin infix operators are redefined
# TODO(coey) handle comparison chains (e.g. 0 < x < y <= 1) or throw error
class Operators(dsl.Context):
    def _supports_binary_op(self, op):
        return op in _builtin_binary_map

    def _make_binary_op(self, op, left, right):
        return _make_fo_expr(_builtin_binary_map[op], left, right)

def operators():
    return Operators(dsl.get_graph())

# Maps for Builtins operator to SolverLib operator ID
_builtin_binary_map = {
    Builtins.plus: 10,
    Builtins.minus: 11,
    Builtins.mult: 12,
    Builtins.div: 13,
    Builtins.pow: 14,
    Builtins.approx_eq: 30,
    Builtins.neq: 31,
    Builtins.lte: 32,
    Builtins.gte: 33,
    Builtins.lt: 34,
    Builtins.gt: 35,
}

# Concatenate arguments into a string separated by underscores
def make_string(args:List):
    string = args[0]
    for arg in args[1:]:
        string = rel_sv.concat(rel_sv.concat(string, "_"), arg)
    return string


# --------------------------------------------------
# Solver
# --------------------------------------------------


@dataclass
class PollingState:
    job_id: str
    continuation_token: str
    is_done: bool
    log_to_console: bool


class Solver:
    def __init__(
        self,
        solver_name: str,
        engine_name: str | None = None,
        resources: SnowflakeResources | None = None,
    ):
        self.provider = Provider(resources=resources)
        self.solver_name = solver_name.lower()

        rai_config = self.provider.resources.config
        settings: dict[str, Any] = {}
        if "experimental" in rai_config:
            exp_config = rai_config.get("experimental", {})
            if isinstance(exp_config, dict):
                if "solvers" in exp_config:
                    settings = exp_config["solvers"].copy()

        # Engine field is not necessary for the solver settings.
        engine_name = engine_name or settings.pop("engine", None)
        if not engine_name:
            engine_name = self.provider.resources.get_user_based_engine_name()

        # The settings are used when creating a solver engine, they
        # may configure each individual solver.
        self.settings = settings

        return self._auto_create_solver_async(engine_name)

    # --------------------------------------------------
    # Helper
    # --------------------------------------------------
    def _auto_create_solver_async(self, name: str):
        with Spinner(
            "Checking solver status",
            leading_newline=True,
        ) as spinner:
            engine = None
            engines = [e for e in self.provider.list_solvers() if e["name"] == name]
            assert len(engines) == 1 or len(engines) == 0
            if len(engines) != 0:
                engine = engines[0]
            if engine:
                # Make sure that the solver requested is enabled
                # on the engine.
                if self.solver_name not in engine["solvers"]:
                    raise Exception(
                        f"Solver `{self.solver_name}` is not enabled on `{name}`."
                    )

                if engine["state"] == "PENDING":
                    spinner.update_messages(
                        {
                            "finished_message": f"Solver {name} is starting",
                        }
                    )
                    pass
                elif engine["state"] == "SUSPENDED":
                    spinner.update_messages(
                        {
                            "finished_message": f"Resuming solver {name}",
                        }
                    )
                    self.provider.resume_solver_async(name)
                elif engine["state"] == "READY":
                    spinner.update_messages(
                        {
                            "finished_message": f"Solver {name} is ready",
                        }
                    )
                    pass
                else:
                    spinner.update_messages(
                        {
                            "message": f"Restarting solver {name}",
                        }
                    )
                    self.provider.delete_solver(name)
                    engine = None
            if not engine:
                # Validate Gurobi config.
                if self.solver_name == "gurobi":
                    is_gurobi_configured = False
                    gurobi_config = self.settings.get("gurobi", {})
                    if all(
                        k in gurobi_config
                        for k in ["license_secret_name", "external_access_integration"]
                    ):
                        is_gurobi_configured = True
                    if not is_gurobi_configured:
                        raise Exception(
                            "Gurobi is not properly configured. You need to provide both `license_secret_name` and `external_access_integration`."
                        )
                self.provider.create_solver_async(name, settings=self.settings)
                spinner.update_messages(
                    {
                        "finished_message": f"Starting solver {name}...",
                    }
                )

            # Fetch the engine object and cache it.
            engine = self.provider.get_solver(name)
            self.engine_cached_at_secs = time.time()
            self.engine = engine

    def _exec_job_async(self, payload):
        payload_json = json.dumps(payload)
        engine_name = self.engine["name"]
        res = self.provider.resources._exec(
            textwrap.dedent(f"""
            CALL {APP_NAME}.experimental.exec_job_async('{ENGINE_TYPE_SOLVER}', '{engine_name}', '{payload_json}')
            """)
        )
        return res[0]["ID"]

    def _exec_job(self, payload, log_to_console=True):
        # We cache the engine object, s.t. we don't have to do a
        # lookup each time we submit a job. We use the auto suspend
        # settings of the engine as a rough estimate on how often we
        # should do this. We want to catch a suspended engine, so that
        # we can resume it below.
        if (time.time() - self.engine_cached_at_secs) >= self.engine["auto_suspend_mins"] * 60:
            self.engine = self.provider.get_solver(self.engine["name"])
            self.engine_cached_at_secs = time.time()

        # Make sure the engine is ready.
        if self.engine["state"] != "READY":
            poll_with_specified_overhead(lambda: self._is_solver_ready(), 0.1)

        with debugging.span("job") as job_span:
            job_id = self._exec_job_async(payload)
            job_span["job_id"] = job_id
            debugging.event("job_created", job_span, job_id=job_id, engine_name=self.engine["name"], job_type=ENGINE_TYPE_SOLVER)
            polling_state = PollingState(job_id, "", False, log_to_console)

            try:
                with debugging.span("wait", job_id=job_id):
                    poll_with_specified_overhead(
                        lambda: self._check_job_status(polling_state), 0.1
                    )
            except KeyboardInterrupt as e:
                print(f"Canceling job {job_id}")
                self.provider.cancel_job(job_id)
                raise e

            return job_id

    def _is_solver_ready(self):
        result = self.provider.get_solver(self.engine["name"])
        self.engine = result
        state = result["state"]
        if state != "READY" and state != "PENDING":
            # Might have suspended or otherwise gone. Recreate.
            self._auto_create_solver_async(self.engine["name"])
        return state == "READY"

    def _check_job_status(self, state):
        response = self.provider.get_job(state.job_id)
        assert response, f"No results from get_job('{state.job_id}')"

        status: str = response["state"]

        self._print_solver_logs(state)

        return status == "COMPLETED" or status == "FAILED" or status == "CANCELED"

    def _get_job_events(self, job_id: str, continuation_token: str = ""):
        results = self.provider.resources._exec(
            f"SELECT {APP_NAME}.experimental.get_job_events('{ENGINE_TYPE_SOLVER}', '{job_id}', '{continuation_token}');"
        )
        if not results:
            return {"events": [], "continuation_token": None}
        row = results[0][0]
        return json.loads(row)

    def _print_solver_logs(self, state: PollingState):
        if state.is_done:
            return

        resp = self._get_job_events(state.job_id, state.continuation_token)

        # Print solver logs to stdout.
        for event in resp["events"]:
            if event["type"] == "LogMessage":
                if state.log_to_console:
                    print(event["event"]["message"])
            else:
                continue

        state.continuation_token = resp["continuation_token"]
        if state.continuation_token == "":
            state.is_done = True


# --------------------------------------------------
# Provider
# --------------------------------------------------
#
# TODO (dba) We use an experimental and unified engine API for
# solvers. Once it is no longer experimental we can remove the
# provider here and use the normal PyRel provider.


class Provider:
    def __init__(self, resources=None):
        if not resources:
            resources = relationalai.Resources()
        if type(resources) is not relationalai.clients.snowflake.Resources:
            raise Exception("Solvers are only supported on SPCS.")
        self.resources = resources

    def create_solver(
        self,
        name: str,
        size: str | None = None,
        settings: dict | None = None,
        auto_suspend_mins: int | None = None,
    ):
        if size is None:
            size = "HIGHMEM_X64_S"
        if settings is None:
            settings = {}
        engine_config: dict[str, Any] = {"settings": settings}
        if auto_suspend_mins is not None:
            engine_config["auto_suspend_mins"] = auto_suspend_mins
        self.resources._exec(
            f"CALL {APP_NAME}.experimental.create_engine('{ENGINE_TYPE_SOLVER}', '{name}', '{size}', {engine_config});"
        )

    def create_solver_async(
        self,
        name: str,
        size: str | None = None,
        settings: dict = {},
        auto_suspend_mins: int | None = None,
    ):
        if size is None:
            size = "HIGHMEM_X64_S"
        if settings is None:
            settings = ""
        engine_config: dict[str, Any] = {"settings": settings}
        if auto_suspend_mins is not None:
            engine_config["auto_suspend_mins"] = auto_suspend_mins
        self.resources._exec(
            f"CALL {APP_NAME}.experimental.create_engine_async('{ENGINE_TYPE_SOLVER}', '{name}', '{size}', {engine_config});"
        )

    def delete_solver(self, name: str):
        self.resources._exec(
            f"CALL {APP_NAME}.experimental.delete_engine('{ENGINE_TYPE_SOLVER}', '{name}');"
        )

    def resume_solver_async(self, name: str):
        self.resources._exec(
            f"CALL {APP_NAME}.experimental.resume_engine_async('{ENGINE_TYPE_SOLVER}', '{name}');"
        )

    def get_solver(self, name: str):
        results = self.resources._exec(
            f"CALL {APP_NAME}.experimental.get_engine('{ENGINE_TYPE_SOLVER}', '{name}');"
        )
        return solver_list_to_dicts(results)[0]

    def list_solvers(self, state: str | None = None):
        where_clause = f"WHERE TYPE='{ENGINE_TYPE_SOLVER}'"
        where_clause = (
            f"{where_clause} AND STATUS = '{state.upper()}'" if state else where_clause
        )
        statement = f"SELECT NAME,ID,SIZE,STATUS,CREATED_BY,CREATED_ON,UPDATED_ON,AUTO_SUSPEND_MINS,SETTINGS FROM {APP_NAME}.experimental.engines {where_clause};"
        results = self.resources._exec(statement)
        return solver_list_to_dicts(results)

    # --------------------------------------------------
    # Job API
    # --------------------------------------------------

    def list_jobs(self, state=None, limit=None):
        state_clause = f"AND STATE = '{state.upper()}'" if state else ""
        limit_clause = f"LIMIT {limit}" if limit else ""
        results = self.resources._exec(
            f"SELECT ID,STATE,CREATED_BY,CREATED_ON,FINISHED_AT,DURATION,PAYLOAD,ENGINE_NAME FROM {APP_NAME}.experimental.jobs where type='{ENGINE_TYPE_SOLVER}' {state_clause} ORDER BY created_on DESC {limit_clause};"
        )
        return job_list_to_dicts(results)

    def get_job(self, id: str):
        results = self.resources._exec(
            f"CALL {APP_NAME}.experimental.get_job('{ENGINE_TYPE_SOLVER}', '{id}');"
        )
        return job_list_to_dicts(results)[0]

    def cancel_job(self, id: str):
        self.resources._exec(
            f"CALL {APP_NAME}.experimental.cancel_job('{ENGINE_TYPE_SOLVER}', '{id}');"
        )


def solver_list_to_dicts(results):
    if not results:
        return []
    return [
        {
            "name": row["NAME"],
            "id": row["ID"],
            "size": row["SIZE"],
            "state": row["STATUS"],  # callers are expecting 'state'
            "created_by": row["CREATED_BY"],
            "created_on": row["CREATED_ON"],
            "updated_on": row["UPDATED_ON"],
            "auto_suspend_mins": row["AUTO_SUSPEND_MINS"],
            "solvers": []
            if row["SETTINGS"] == ""
            else [
                k
                for (k, v) in json.loads(row["SETTINGS"]).items()
                if isinstance(v, dict) and v.get("enabled", False)
            ],
        }
        for row in results
    ]


def job_list_to_dicts(results):
    if not results:
        return []
    return [
        {
            "id": row["ID"],
            "state": row["STATE"],
            "created_by": row["CREATED_BY"],
            "created_on": row["CREATED_ON"],
            "finished_at": row["FINISHED_AT"],
            "duration": row["DURATION"] if "DURATION" in row else 0,
            "solver": json.loads(row["PAYLOAD"])["solver"],
            "engine": row["ENGINE_NAME"],
        }
        for row in results
    ]
